import logging
import os

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import ValidationState, Buffer
from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import Condition, completion_is_selected, is_searching, has_completions
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.key_processor import KeyPressEvent

from peek.common import HTTP_METHODS
from peek.errors import PeekSyntaxError, PeekError
from peek.visitors import FormattingVisitor

_logger = logging.getLogger(__name__)


def key_bindings(app):
    kb = KeyBindings()

    @kb.add('tab', filter=~(completion_is_selected | is_searching | has_completions))
    def _(event):
        event.current_buffer.insert_text('  ')

    @kb.add('enter', filter=~(completion_is_selected | is_searching) & buffer_should_be_handled(app))
    def _(event):
        event.current_buffer.validate_and_handle()
        if app.capture.file() is not None:
            print(event.current_buffer.text, file=app.capture.file())

    @kb.add('enter', filter=~(completion_is_selected | is_searching) & ~buffer_should_be_handled(app))
    def _(event):
        b = event.current_buffer  # type: Buffer
        c = b.document.current_char
        _logger.debug(f'not handling enter, current char {c!r}')

        # When cursor is on the right curly bracket, just use the ident of the current line
        existing_indent = 0
        for x in b.document.current_line:
            if x.isspace():
                existing_indent += 1
            else:
                break
        if c == '}':
            b.insert_text('\n' + ' ' * (existing_indent + 2))
            b.insert_text('\n' + ' ' * existing_indent, move_cursor=False)
        else:
            b.insert_text('\n' + ' ' * existing_indent)

    @kb.add('escape', 'enter', filter=~(completion_is_selected | is_searching))
    def _(event):
        event.app.current_buffer.newline()

    @kb.add('escape', 'c')
    def _(event: KeyPressEvent):
        get_app().clipboard.set_text(event.app.current_buffer.text)
        app.preserved_text = event.current_buffer.text
        event.current_buffer.reset()
        event.current_buffer.validate_and_handle()

    @kb.add('c-d')
    def _(event):
        app.signal_exit()
        # Short circuit the validation
        event.current_buffer.validation_state = ValidationState.VALID
        event.current_buffer.validate_and_handle()

    @kb.add('f3')
    def _(event):
        _logger.debug('Reformatting payload json')
        try:
            texts = []
            nodes = app.parser.parse(event.current_buffer.text)
            for j, node in enumerate(nodes):
                texts.append(FormattingVisitor(pretty=app.is_pretty).visit(node))
            event.current_buffer.text = '\n'.join(texts)
            app.is_pretty = not app.is_pretty
        except PeekSyntaxError as e:
            _logger.debug(f'Cannot reformat for invalid/incomplete input: {e}')

    def switch_connection(event):
        _logger.debug(f'switching to connection: {event.key_sequence[1].key}')
        try:
            app.es_client_manager.set_current(int(event.key_sequence[1].key))
            app.preserved_text = event.current_buffer.text
            event.current_buffer.reset()
            event.current_buffer.validate_and_handle()
        except PeekError as e:
            app.display.error(e)

    for i in range(5):
        kb.add('escape', f'{i}')(switch_connection)

    # par-editing is a shamelessly copy from https://github.com/nicolewhite/cycli/blob/master/cycli/binder.py
    @kb.add("{")
    def curly_left(event):
        b = event.current_buffer
        b.insert_text("{")
        b.insert_text("}", move_cursor=False)

    @kb.add("}")
    def curly_right(event):
        b = event.current_buffer
        char = b.document.current_char

        if char == "}":
            b.cursor_right()
        else:
            b.insert_text("}")

    @kb.add("(")
    def paren_left(event):
        b = event.current_buffer
        b.insert_text("(")
        b.insert_text(")", move_cursor=False)

    @kb.add(")")
    def paren_right(event):
        b = event.current_buffer
        char = b.document.current_char

        if char == ")":
            b.cursor_right()
        else:
            b.insert_text(")")

    @kb.add("[")
    def bracket_left(event):
        b = event.current_buffer
        b.insert_text("[")
        b.insert_text("]", move_cursor=False)

    @kb.add("]")
    def bracket_right(event):
        b = event.current_buffer
        char = b.document.current_char

        if char == "]":
            b.cursor_right()
        else:
            b.insert_text("]")

    @kb.add("'")
    def apostrophe(event):
        b = event.current_buffer
        char = b.document.current_char

        if char == "'":
            b.cursor_right()
        else:
            b.insert_text("'")
            b.insert_text("'", move_cursor=False)

    @kb.add("\"")
    def quote(event):
        b = event.current_buffer
        char = b.document.current_char

        if char == "\"":
            b.cursor_right()
        else:
            b.insert_text("\"")
            b.insert_text("\"", move_cursor=False)

    @kb.add("`")
    def backtick(event):
        b = event.current_buffer
        char = b.document.current_char

        if char == "`":
            b.cursor_right()
        else:
            b.insert_text("`")
            b.insert_text("`", move_cursor=False)

    @kb.add('c-h')  # backspace
    def backspace(event):
        b = event.current_buffer
        current_char = b.document.current_char
        before_char = b.document.char_before_cursor

        patterns = [("(", ")"), ("[", "]"), ("{", "}"), ("'", "'"), ('"', '"'), ("`", "`")]

        for pattern in patterns:
            if before_char == pattern[0] and current_char == pattern[1]:
                b.cursor_right()
                b.delete_before_cursor(2)
                return

        b.delete_before_cursor()

    return kb


def buffer_should_be_handled(app):
    @Condition
    def cond():
        doc = get_app().layout.get_buffer_by_name(DEFAULT_BUFFER).document
        _logger.debug(f'current doc: {doc}')
        # Always handle empty text
        if doc.text.strip() == '':
            return True

        if doc.empty_line_count_at_the_end() + 1 == doc.line_count and doc.line_count > 1:
            _logger.debug('Lines are all blank after the first one')
            return True

        # Do not handle if there are chars after the cursor position
        if doc.text[doc.cursor_position:].strip() != '':
            _logger.debug('Extra non-white chars found after cursor')
            return False

        # Handle if extra blank line found
        last_linesep_position = doc.text.rfind(os.linesep)
        if last_linesep_position != -1 and doc.text[last_linesep_position:].strip() == '':
            return True

        # If there are existing lines already, do not handle
        if doc.text.count('\n') > 0:
            _logger.debug('Existing line found above current line')
            return False
        else:  # If this is the 1st line, do not handle if it is ES API call
            return doc.text.lstrip().split(maxsplit=1)[0].lower() not in (HTTP_METHODS + ['for'])

    return cond



import datetime
import sqlite3
from os.path import expanduser
from typing import Iterable, List

from prompt_toolkit.history import History

from peek.config import config_location, ensure_dir_exists

HIST_MAX = 10_000


class SqLiteHistory(History):

    def __init__(self, history_max=HIST_MAX):
        super().__init__()
        self.history_max = history_max
        db_file = expanduser(config_location() + 'history')
        ensure_dir_exists(db_file)
        self.conn = sqlite3.connect(db_file)
        self.conn.execute('CREATE TABLE IF NOT EXISTS history '
                          '(id INTEGER PRIMARY KEY AUTOINCREMENT, content TEXT NOT NULL, timestamp INTEGER NOT NULL)')
        self._maintain_size()
        self.conn.commit()

    def __del__(self):
        self.conn.close()

    def _maintain_size(self):
        c = self.conn.cursor()
        res = c.execute('SELECT COUNT(*) from history').fetchone()
        if res is None or res[0] < self.history_max:
            return
        c.execute('DELETE FROM history where id in (SELECT id FROM history ORDER BY id limit ?)',
                  (res[0] - self.history_max,))

    def load_history_strings(self) -> Iterable[str]:
        strings: List[str] = []
        for row in self.conn.execute('SELECT * FROM history ORDER BY id DESC'):
            strings.append(row[1])
        return strings

    def store_string(self, string: str) -> None:
        self.conn.execute("INSERT INTO history(content, timestamp) VALUES (?, ?)", (string, datetime.datetime.now()))
        self.conn.commit()

    def load_recent(self, size=100):
        lines = []
        for row in self.conn.execute("select * from history where id > (select max(id) from history) - ?", (size,)):
            lines.append((row[0], row[1]))
        return lines

    def get_entry(self, index):
        if index > 0:
            for row in self.conn.execute('select * from history where id = ?', (index,)):
                return row[0], row[1]
            else:
                return None
        elif index < 0:
            for row in self.conn.execute('select * from history where id = (select max(id) from history) + ?',
                                         (index,)):
                return row[0], row[1]
            else:
                return None
        else:
            return None



import functools
import logging

from peek.ast import Visitor, EsApiCallNode, DictNode, KeyValueNode, ArrayNode, NumberNode, \
    StringNode, FuncCallNode, NameNode, TextNode, ShellOutNode, EsApiCallInlinePayloadNode, EsApiCallFilePayloadNode, \
    BinOpNode, UnaryOpNode, GroupNode, SymbolNode, LetNode, ForInNode

_logger = logging.getLogger(__name__)


class Ref:

    def __init__(self, init_value=None):
        self._value = init_value

    def set(self, new_value):
        self._value = new_value

    def get(self):
        return self._value


class FormattingVisitor(Visitor):
    def __init__(self, pretty=False):
        super().__init__()
        self.pretty = pretty
        self.parts = []
        self.indent_level = 0

    def visit(self, node):
        self.parts = []
        self.indent_level = 0
        with self.consumer(lambda *args: self.parts.extend(args)):
            node.accept(self)
        return ''.join(self.parts)

    def visit_es_api_call_node(self, node: EsApiCallNode):
        self.consume(node.method_node.token.value, ' ')
        if isinstance(node.path_node, TextNode):
            self.consume(node.path_node.token.value)
        else:
            node.path_node.accept(self)
        options_parts = []
        options_consumer = functools.partial(options_consumer_maker, options_parts)
        self._do_visit_dict_node(node.options_node, options_consumer)
        if options_parts:
            self.consume(' ' + ''.join(options_parts))
        self.consume('\n')
        self.consume(self._indent())

        if isinstance(node, EsApiCallInlinePayloadNode):
            for dict_node in node.dict_nodes:
                dict_node.accept(self)
                self.consume('\n')
        elif isinstance(node, EsApiCallFilePayloadNode):
            node.file_node.accept(self)
            self.consume('\n')
        else:
            raise ValueError(f'Unknown node: {node!r}')

    def visit_func_call_node(self, node: FuncCallNode):
        node.name_node.accept(self)
        if node.is_stmt:
            self.consume(' ')
        else:
            self.consume('(')

        arg_separator = self._arg_separator(node)

        symbols_parts = []

        def func_symbols_consumer(*args):
            for v in args:
                if v == ',':
                    symbols_parts.append(arg_separator)
                elif v not in ['[', ']'] and v.strip() != '':
                    symbols_parts.append(f'{v}')

        self._do_visit_array_node(node.symbols_node, func_symbols_consumer)
        if symbols_parts:
            self.consume(''.join(symbols_parts))

        args_parts = []

        def func_args_consumer(*args):
            for v in args:
                if v == ',':
                    args_parts.append(arg_separator)
                elif v not in ['[', ']'] and v.strip() != '':
                    args_parts.append(v)

        self._do_visit_array_node(node.args_node, func_args_consumer)

        if args_parts:
            if symbols_parts:
                self.consume(arg_separator)
            self.consume(''.join(args_parts))

        kwargs_parts = []
        func_kwargs_consumer = functools.partial(options_consumer_maker, kwargs_parts)
        self._do_visit_dict_node(node.kwargs_node, func_kwargs_consumer)
        if kwargs_parts:
            kwargs_parts = [x if x != ' ' else arg_separator for x in kwargs_parts]
            if args_parts or args_parts:
                self.consume(arg_separator)
            self.consume(''.join(kwargs_parts))
        if not node.is_stmt:
            self.consume(')')

    def visit_let_node(self, node: LetNode):
        self.consume('let', ' ')
        assignments_parts = []
        assignments_consumer = functools.partial(options_consumer_maker, assignments_parts)
        self._do_visit_dict_node(node.assignments_node, assignments_consumer)
        if assignments_parts:
            self.consume(''.join(assignments_parts))

    def visit_shell_out_node(self, node: ShellOutNode):
        self.consume(f'!{node.command}')

    def visit_for_in_node(self, node: ForInNode):
        self.consume('for', ' ')
        node.item.accept(self)
        self.consume(' ', 'in', ' ')
        node.items.accept(self)
        self.consume(' ', '{', '\n')
        self.indent_level += 1
        for n in node.suite:
            self.consume(self._indent())
            n.accept(self)
            self.consume('\n')
        self.indent_level -= 1
        self.consume(self._indent(), '}')

    def visit_string_node(self, node: StringNode):
        self.consume(node.token.value)

    def visit_number_node(self, node: NumberNode):
        self.consume(node.token.value)

    def visit_name_node(self, node: NameNode):
        self.consume(node.token.value)

    def visit_symbol_node(self, node: SymbolNode):
        self.consume('@', node.token.value)

    def visit_text_node(self, node: TextNode):
        self.consume(node.token.value)

    def visit_key_value_node(self, node: KeyValueNode):
        node.key_node.accept(self)
        self.consume(':')
        if self.pretty:
            self.consume(' ')
        node.value_node.accept(self)

    def visit_dict_node(self, node: DictNode):
        parts = []
        self._do_visit_dict_node(node, lambda v: parts.append(v))
        self.consume(''.join(parts))

    def visit_array_node(self, node: ArrayNode):
        parts = []
        self._do_visit_array_node(node, lambda v: parts.append(v))
        self.consume(''.join(parts))

    def visit_bin_op_node(self, node: BinOpNode):
        node.left_node.accept(self)
        self.consume(node.op_token.value)
        node.right_node.accept(self)

    def visit_unary_op_node(self, node: UnaryOpNode):
        self.consume(node.op_token.value)
        node.operand_node.accept(self)

    def visit_group_node(self, node: GroupNode):
        self.consume('(')
        node.grouped.accept(self)
        self.consume(')')

    def _do_visit_dict_node(self, node: DictNode, consumer):
        self.push_consumer(consumer)
        self.consume('{')
        if node.kv_nodes and self.pretty:
            self.consume('\n')
            self.indent_level += 1
        for i, kv_node in enumerate(node.kv_nodes):
            if self.pretty:
                self.consume(self._indent())
            kv_node.accept(self)
            if i < len(node.kv_nodes) - 1:
                self.consume(',')
                if self.pretty:
                    self.consume('\n')
        if node.kv_nodes and self.pretty:
            self.consume('\n')
            self.indent_level -= 1
            self.consume(self._indent())
        self.consume('}')
        self.pop_consumer()

    def _do_visit_array_node(self, node: ArrayNode, consumer):
        self.push_consumer(consumer)
        self.consume('[')
        if node.value_nodes and self.pretty:
            self.consume('\n')
            self.indent_level += 1
        for i, value_node in enumerate(node.value_nodes):
            self.consume(self._indent())
            value_node.accept(self)
            if i < len(node.value_nodes) - 1:
                self.consume(',')
                if self.pretty:
                    self.consume('\n')
        if node.value_nodes and self.pretty:
            self.consume('\n')
            self.indent_level -= 1
            self.consume(self._indent())
        self.consume(']')
        self.pop_consumer()

    def _indent(self):
        return '  ' * self.indent_level

    def _arg_separator(self, node: FuncCallNode):
        if node.is_stmt:
            return ' '
        else:
            if self.pretty:
                return ', '
            else:
                return ','


class TreeFormattingVisitor(Visitor):

    def __init__(self, indent_chars='  '):
        super().__init__()
        self.indent_chars = indent_chars
        self.lines = []
        self.indent_level = 0
        self.lines = []

    def visit(self, node):
        self.lines = []
        self.push_consumer(lambda v: self.lines.append(v))
        node.accept(self)
        self.pop_consumer()
        return '\n'.join(self.lines)

    def visit_es_api_call_node(self, node: EsApiCallNode):
        self.consume(f'{self._indent()}EsApiStmt')
        self.indent_level += 1
        node.method_node.accept(self)
        node.path_node.accept(self)
        node.options_node.accept(self)
        if isinstance(node, EsApiCallInlinePayloadNode):
            for n in node.dict_nodes:
                n.accept(self)
        elif isinstance(node, EsApiCallFilePayloadNode):
            node.file_node.accept(self)
        else:
            raise ValueError(f'Unknown node: {node!r}')
        self.indent_level -= 1

    def visit_func_call_node(self, node: FuncCallNode):
        self.consume(f'{self._indent()}Func{"Stmt" if node.is_stmt else "Expr"}({node.name_node!r})')
        self.indent_level += 1
        node.symbols_node.accept(self)
        node.args_node.accept(self)
        node.kwargs_node.accept(self)
        self.indent_level -= 1

    def visit_let_node(self, node: LetNode):
        self.consume(f'{self._indent()}LetStmt')
        self.indent_level += 1
        node.assignments_node.accept(self)
        self.indent_level -= 1

    def visit_shell_out_node(self, node: ShellOutNode):
        self.consume(f'{self._indent()}ShellOut')
        self.indent_level += 1
        node.text_node.accept(self)
        self.indent_level -= 1

    def visit_for_in_node(self, node: ForInNode):
        self.consume(f'{self._indent()}ForIn')
        self.indent_level += 1
        node.item.accept(self)
        node.items.accept(self)
        for n in node.suite:
            n.accept(self)
        self.indent_level -= 1

    def visit_name_node(self, node):
        self.consume(f'{self._indent()}{node.token.value}')

    def visit_symbol_node(self, node):
        self.consume(f'{self._indent()}@{node.token.value}')

    def visit_text_node(self, node):
        self.consume(f'{self._indent()}{node.token.value}')

    def visit_key_value_node(self, node: KeyValueNode):
        self.consume(f'{self._indent()}KV')
        self.indent_level += 1
        node.key_node.accept(self)
        node.value_node.accept(self)
        self.indent_level -= 1

    def visit_dict_node(self, node: DictNode):
        self.consume(f'{self._indent()}Dict')
        self.indent_level += 1
        for kv_node in node.kv_nodes:
            kv_node.accept(self)
        self.indent_level -= 1

    def visit_array_node(self, node: ArrayNode):
        self.consume(f'{self._indent()}Array')
        self.indent_level += 1
        for n in node.value_nodes:
            n.accept(self)
        self.indent_level -= 1

    def visit_string_node(self, node):
        self.consume(f'{self._indent()}{node.token.value}')

    def visit_number_node(self, node):
        self.consume(f'{self._indent()}{node.token.value}')

    def visit_bin_op_node(self, node: BinOpNode):
        self.consume(f'{self._indent()}BinOp({node.op_token.value})')
        self.indent_level += 1
        node.left_node.accept(self)
        node.right_node.accept(self)
        self.indent_level -= 1

    def visit_unary_op_node(self, node: UnaryOpNode):
        self.consume(f'{self._indent()}UnaryOp({node.op_token.value})')
        self.indent_level += 1
        node.operand_node.accept(self)
        self.indent_level -= 1

    def visit_group_node(self, node):
        node.grouped.accept(self)

    def _indent(self):
        return self.indent_level * self.indent_chars


def options_consumer_maker(options_parts, v):
    """
    This is to convert a dict representation to k=v representation
    """
    if v == ':':
        options_parts.append('=')
    elif v == ',':
        options_parts.append(' ')
    elif v not in ['{', '}'] and v.strip() != '':
        options_parts.append(v)



import json
import logging
import urllib

from peek.connection import EsClient, RefreshingEsClient
from peek.errors import PeekError

_logger = logging.getLogger(__name__)


def krb_authenticate(es_client: EsClient, service=None, username=None, name=None):
    _logger.debug(f'Connection with Kerberos: {service}')
    if service is None:
        raise PeekError('Service is required for kerberos authentication')
    import kerberos
    result, context = kerberos.authGSSClientInit(service, principal=username)
    kerberos.authGSSClientStep(context, '')
    ticket = kerberos.authGSSClientResponse(context)
    _logger.debug(f'Kerberos ticket: {ticket}')
    auth_response = es_client.perform_request(
        'POST',
        '/_security/oauth2/token',
        json.dumps({
            'grant_type': '_kerberos',
            'kerberos_ticket': ticket,
        }),
        deserialize_it=True)
    _logger.debug(f'Kerberos Token auth response: {auth_response}')

    return RefreshingEsClient(
        es_client,
        '_KRB',
        auth_response['access_token'],
        auth_response['refresh_token'],
        auth_response['expires_in'],
        name=name)


class KrbAuthenticateFunc:
    def __call__(self, app, **options):
        service = options.get('service', None)
        conn = options.get('conn', None)
        if service is None:
            if app.es_client_manager.current.hosts:
                host = urllib.parse.urlparse(app.es_client_manager.current.hosts.split(',')[0]).hostname
                service = f'HTTP@{host}'
            else:
                raise PeekError('Cannot infer service principal. Please specify explicitly')

        krb_es_client = krb_authenticate(
            app.es_client_manager.current if conn is None else app.es_client_manager.get_client(conn),
            service,
            options.get('username', None),
            options.get('name', None)
        )
        app.es_client_manager.add(krb_es_client)
        return app.es_client_manager.current.perform_request('GET', '/_security/_authenticate')

    @property
    def options(self):
        return {'service': '', 'username': '', 'name': None, 'conn': None}

    @property
    def description(self):
        return 'Start Kerberos authentication flow'



#!/usr/bin/env python

"""Tests for `peek` package."""

from peek.peekapp import PeekApp


def test_multiple_stmts():
    text = """get abc

post abc/_doc
{ "foo":
         "bar"
}

conn foo=bar
get abc
"""
    nodes = []
    peek = PeekApp(batch_mode=True, extra_config_options=('log_level=None',))
    peek.execute_node = lambda stmt: nodes.append(stmt)
    peek.process_input(text)

    for n in nodes:
        print(n)

    assert [str(n) for n in nodes] == [
        r'''get abc {}
''',
        r'''post abc/_doc {}
{"foo":"bar"}
''',
        r'''conn [] [] {foo:bar}
''',
        r'''get abc {}
''',
    ]



from unittest.mock import MagicMock, call

import pytest
from configobj import ConfigObj

from peek.parser import PeekParser
from peek.vm import PeekVM


@pytest.fixture
def peek_vm():
    mock_app = MagicMock(name='PeekApp')

    vm = PeekVM(mock_app)
    mock_app.vm = vm
    mock_app.config = ConfigObj()
    vm.context['debug'] = MagicMock(return_value=None)
    mock_app.display.info = MagicMock(return_value=None)
    mock_app.display.error = MagicMock(return_value=None)
    es_client = MagicMock(name='EsClient')
    mock_app.es_client_manager.current = es_client
    mock_app.es_client_manager.get_client = MagicMock(return_value=es_client)
    es_client.perform_request = MagicMock(return_value='{"foo": [1, 2, 3, 4], "bar": {"hello": [42, "world"]}}')
    return vm


@pytest.fixture
def parser():
    return PeekParser()


def assert_called_with(vm, *args, **kwargs):
    vm.context['debug'].assert_called_with(vm.app, *args, **kwargs)


def test_peek_vm_data_types(peek_vm, parser):
    peek_vm.execute_node(parser.parse('debug "42" 42 true 4.2 [4, 2] {4:2} @a42 foo="bar"')[0])
    assert_called_with(peek_vm, '42', 42, True, 4.2, [4, 2], {4: 2}, **{'@': ['a42'], 'foo': 'bar'})


def test_peek_vm_expr(peek_vm, parser):
    peek_vm.execute_node(parser.parse('debug 1 + 2 * 3 + (5-1) {"a": 42}.@a [4, 2, 1].1 foo="a" + "b"')[0])
    assert_called_with(peek_vm, 11, 42, 2, **{'foo': 'ab'})


def test_str_and_number(peek_vm, parser):
    peek_vm.execute_node(parser.parse('debug "hello" + 42')[0])
    assert_called_with(peek_vm, 'hello42')


def test_peek_vm_func(peek_vm, parser):
    peek_vm.execute_node(parser.parse('debug echo(0 1 2)')[0])
    assert_called_with(peek_vm, '0 1 2')
    peek_vm.execute_node(parser.parse('debug [echo].0(0 1 2)')[0])
    assert_called_with(peek_vm, '0 1 2')


def test_peek_vm_es_api_call(peek_vm, parser):
    peek_vm.execute_node(parser.parse('GET /')[0])
    assert peek_vm.get_value('_') == {'foo': [1, 2, 3, 4], 'bar': {'hello': [42, 'world']}}
    peek_vm.execute_node(parser.parse('debug _."bar".@hello.1')[0])
    assert_called_with(peek_vm, 'world')
    peek_vm.execute_node(parser.parse('GET ("foo" + "/" + 42)')[0])
    peek_vm.app.es_client_manager.current.perform_request.assert_called_with('GET', '/foo/42', None, headers=None)


def test_peek_vm_let(peek_vm, parser):
    peek_vm.execute_node(parser.parse('let foo={"a": [3, 4, 5]} bar=["hello", 42] c="world"')[0])
    peek_vm.execute_node(parser.parse('debug foo bar c')[0])
    assert_called_with(peek_vm, {"a": [3, 4, 5]}, ["hello", 42], "world")
    peek_vm.execute_node(parser.parse('let foo.@a.1 = 42')[0])
    peek_vm.execute_node(parser.parse('debug foo')[0])
    assert_called_with(peek_vm, {"a": [3, 42, 5]})


def test_for_in(peek_vm, parser):
    peek_vm.execute_node(parser.parse('''for x in [1, 2, 3] {
    let y = x
    debug x
}''')[0])

    assert peek_vm.get_value('y') == 3
    peek_vm.context['debug'].assert_has_calls(
        [call(peek_vm.app, 1), call(peek_vm.app, 2), call(peek_vm.app, 3)])



import logging

from configobj import ConfigObj

from peek.config import config_location
from peek.connection import ConnectFunc
from peek.errors import PeekError
from peek.krb import KrbAuthenticateFunc
from peek.oidc import OidcAuthenticateFunc
from peek.saml import SamlAuthenticateFunc

_logger = logging.getLogger(__name__)


class ConfigFunc:

    def __call__(self, app, **options):
        if not options:
            return {
                'location': config_location(),
                'config': app.config
            }

        extra_config = {}
        for key, value in options.items():
            parent = extra_config
            key_components = key.split('.')
            for key_component in key_components[:-1]:
                child = parent.get(key_component)
                if child is None:
                    parent[key_component] = {}
                elif not isinstance(child, dict):
                    _logger.warning(f'Config key [{key}] conflicts. '
                                    f'Value of [{key_component}] is not a [dict], '
                                    f'but [{type(child)}]')
                    parent = None
                    break
                parent = parent[key_component]

            if isinstance(parent, dict):
                parent[key_components[-1]] = value

        # TODO: saner merge that does not change data type, e.g. from dict to primitive and vice versa
        app.config.merge(ConfigObj(extra_config))

    @property
    def description(self):
        return 'View and set config options'


class SessionFunc:

    def __call__(self, app, current=None, **options):
        for symbol in options.get('@', []):
            if symbol == 'info':
                return app.es_client_manager.current.info()
            elif symbol == 'remove':
                app.es_client_manager.remove_client(None)  # remove current client
            else:
                raise PeekError(f'Unknown sub-command: {symbol!r}')

        current = current if current is not None else options.get('current', None)
        if current is not None:
            app.es_client_manager.set_current(current)

        remove = options.get('remove', None)
        if remove is not None:
            app.es_client_manager.remove_client(remove)

        rename = options.get('rename', None)
        if rename:
            app.es_client_manager.current.name = str(rename)

        info = options.get('info', None)
        if info is not None:
            return app.es_client_manager.get_client(info).info()

        return str(app.es_client_manager)

    @property
    def options(self):
        return {'current': None, 'remove': None, 'rename': None, 'info': None,
                '@info': None, '@remove': None}

    @property
    def description(self):
        return 'List sessions and set current session'


class RunFunc:

    def __call__(self, app, file, **options):
        with open(file) as ins:
            app.process_input(ins.read(), echo=options.get('echo', True))

    @property
    def options(self):
        return {'echo': False}

    @property
    def description(self):
        return 'Load and execute external script'


class ResetFunc:

    def __call__(self, app, *args, **kwargs):
        app.reset()

    @property
    def description(self):
        return 'Reset state of the program'


class HistoryFunc:

    def __call__(self, app, index=None, **options):
        if index is None:
            history = []
            for entry in app.history.load_recent(size=options.get('size', 100)):
                history.append(f'{entry[0]:>6} {entry[1]!r}')
            return '\n'.join(history)
        else:
            entry = app.history.get_entry(index)
            if entry is None:
                raise PeekError(f'History not found for index: {index}')
            app.process_input(entry[1])

    @property
    def options(self):
        return {'size': 100}

    @property
    def description(self):
        return 'View history and execute by history id'


class RangeFunc:

    def __call__(self, app, start, stop, step=1):
        return list(range(start, stop, step))

    @property
    def description(self):
        return 'Range over given start and stop (exclusive) and optionally a step'


class EchoFunc:

    def __call__(self, app, *args, **options):
        return ' '.join(str(arg) for arg in args)

    @property
    def description(self):
        return 'Print given items'


class CaptureFunc:

    def __call__(self, app, f=None, **options):
        directives = options.get('@')
        if not directives:
            return app.capture.status()

        # Only honor first directive
        directive = directives[0]
        if directive == 'start':
            return app.start_capture(f)

        elif directive == 'stop':
            return app.stop_capture()
        else:
            raise PeekError(f'Unknown capture directive: {directive}')

    @property
    def options(self):
        return {'@start': None, '@stop': None}

    @property
    def description(self):
        return 'Capture session'


class HelpFunc:

    def __call__(self, app, func=None, **options):
        if func is None:
            return '\n'.join(k for k in app.vm.functions.keys())

        for k, v in app.vm.functions.items():
            if v == func:
                description = getattr(v, "description", None)
                header = k + (f' - {description}' if description else '')
                return f'{header}\n{getattr(func, "options", {})}'
        else:
            raise PeekError(f'No such function: {func}')

    @property
    def description(self):
        return 'List available functions and show help message of a function'


EXPORTS = {
    'connect': ConnectFunc(),
    'config': ConfigFunc(),
    'session': SessionFunc(),
    'run': RunFunc(),
    'history': HistoryFunc(),
    'echo': EchoFunc(),
    'range': RangeFunc(),
    'reset': ResetFunc(),
    'capture': CaptureFunc(),
    'help': HelpFunc(),
    'saml_authenticate': SamlAuthenticateFunc(),
    'oidc_authenticate': OidcAuthenticateFunc(),
    'krb_authenticate': KrbAuthenticateFunc(),
}



import pytest

from peek.parser import PeekParser
from peek.visitors import FormattingVisitor, TreeFormattingVisitor


@pytest.fixture
def parser():
    return PeekParser()


def test_formatting_api_call_compact(parser):
    visitor = FormattingVisitor()

    nodes = parser.parse('''get / conn =  1 runas = "foo"
{
  "a"   :
"b", "q": [
1,
2]}''')

    assert visitor.visit(nodes[0]) == '''get / conn=1 runas="foo"
{"a":"b","q":[1,2]}
'''
    nodes = parser.parse('get ("foo" + "/" + 42 + "/" + 1)')
    assert visitor.visit(nodes[0]) == 'get ("foo"+"/"+42+"/"+1)\n'


def test_formatting_api_call_pretty(parser):
    visitor = FormattingVisitor(pretty=True)
    nodes = parser.parse('''get / conn =  1 runas = "foo"
{
  "a"   :
"b", "q": [
1,
2]}''')

    assert visitor.visit(nodes[0]) == '''get / conn=1 runas="foo"
{
  "a": "b",
  "q": [
    1,
    2
  ]
}
'''


def test_formatting_func_call(parser):
    visitor = FormattingVisitor()
    nodes = parser.parse('''f @abc 1 "a" b=a foo="bar"''')

    assert visitor.visit(nodes[0]) == '''f @abc 1 "a" b=a foo="bar"'''


def test_formatting_for_in_compact(parser):
    visitor = FormattingVisitor(pretty=False)
    nodes = parser.parse('''for a in [1, 2, 3] {
        echo a
        for b in c {
            echo b
            let x = b
        }
        echo a + 1
    }''')

    assert visitor.visit(nodes[0]) == '''for a in [1,2,3] {
  echo a
  for b in c {
    echo b
    let x=b
  }
  echo a+1
}'''


def test_formatting_pretty(parser):
    visitor = FormattingVisitor(pretty=True)
    nodes = parser.parse('''for x in [1,2] {
    let a = { 1: 2, 3: 4, 'foo': [42, "hello"]}
    echo a x
    for y in c {
        echo [1, 3, 5] b={'foo': 42, "bar": [6, 7, [8, 9]]}
        GET /path
        {"foo": {"q": [1, 2,{'x': [3, "5", {}]}]}}
    }
}''')
    assert visitor.visit(nodes[0]) == '''for x in [
  1,
  2
] {
  let a={
      1: 2,
      3: 4,
      'foo': [
        42,
        "hello"
      ]
    }
  echo a x
  for y in c {
    echo [
        1,
        3,
        5
      ] b={
        'foo': 42,
        "bar": [
          6,
          7,
          [
            8,
            9
          ]
        ]
      }
    GET /path
    {
      "foo": {
        "q": [
          1,
          2,
          {
            'x': [
              3,
              "5",
              {}
            ]
          }
        ]
      }
    }

  }
}'''


def test_formatting_looped_api_calls(parser):
    nodes = parser.parse('''for i in range(@foo @bar 1 10 a=4 b=2) {
  POST test-1/_doc
  { "tag": i }
  POST test-2/_doc
  { "tag": i + 1}
}
''')

    assert FormattingVisitor(pretty=True).visit(nodes[0]) == '''for i in range(@foo, @bar, 1, 10, a=4, b=2) {
  POST test-1/_doc
  {
    "tag": i
  }

  POST test-2/_doc
  {
    "tag": i+1
  }

}'''

    assert FormattingVisitor(pretty=False).visit(nodes[0]) == '''for i in range(@foo,@bar,1,10,a=4,b=2) {
  POST test-1/_doc
  {"tag":i}

  POST test-2/_doc
  {"tag":i+1}

}'''


def test_tree_formatting(parser):
    visitor = TreeFormattingVisitor()
    nodes = parser.parse('''f @abc 1 "a" b=a foo="bar" 1 * 2 + (3-2) * 5 / 6
get ("hello" + 42 + "world")''')
    assert visitor.visit(nodes[0]) == '''FuncStmt('f')
  Array
    @abc
  Array
    1
    "a"
    BinOp(+)
      BinOp(*)
        1
        2
      BinOp(/)
        BinOp(*)
          BinOp(-)
            3
            2
          5
        6
  Dict
    KV
      b
      a
    KV
      foo
      "bar"'''

    assert visitor.visit(nodes[1]) == '''EsApiStmt
  get
  BinOp(+)
    BinOp(+)
      "hello"
      42
    "world"
  Dict'''


def test_tree_formatting_func_expr_chain(parser):
    visitor = TreeFormattingVisitor()
    nodes = parser.parse('''f a.@b(1)''')
    assert "FuncExpr('a . b')" in visitor.visit(nodes[0])


def test_tree_formatting_for_in(parser):
    visitor = TreeFormattingVisitor()
    nodes = parser.parse('''for a in [1, 2, 3] {
    echo a
    for b in c {
        echo b
    }
    echo a + 1
}
for x in y {}''')
    assert visitor.visit(nodes[0]) == '''ForIn
  a
  Array
    1
    2
    3
  FuncStmt('echo')
    Array
    Array
      a
    Dict
  ForIn
    b
    c
    FuncStmt('echo')
      Array
      Array
        b
      Dict
  FuncStmt('echo')
    Array
    Array
      BinOp(+)
        a
        1
    Dict'''

    assert visitor.visit(nodes[1]) == '''ForIn
  x
  y'''



import json
import logging
import os
import ssl
import time
import webbrowser
from http.server import HTTPServer, BaseHTTPRequestHandler
from threading import Thread
from typing import Any, Optional

from peek.connection import EsClient, RefreshingEsClient

_logger = logging.getLogger(__name__)


class _OidcExchange:
    callback_path = None


class CallbackHTTPRequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        _OidcExchange.callback_path = self.path
        self.send_response(200)
        self.end_headers()
        self.wfile.write(b'Callback received, you can now close the browser tab.')

    def do_POST(self):
        pass

    def log_message(self, fmt: str, *args: Any) -> None:
        _logger.info("%s - - [%s] %s\n" %
                     (self.address_string(),
                      self.log_date_time_string(),
                      fmt % args))


def oidc_authenticate(es_client: EsClient, realm: str, callback_port: str, name: Optional[str]):
    _logger.info(f'OIDC authenticate for realm {realm!r} and callback port {callback_port!r}')
    prepare_response = _oidc_prepare(es_client, realm)
    httpd = _oidc_start_http_server(callback_port)
    print('Please use browser to complete authentication against the idP')
    webbrowser.open(prepare_response['redirect'])
    while _OidcExchange.callback_path is None:
        time.sleep(0.1)
    try:
        httpd.shutdown()
        auth_response = _oidc_do_authenticate(es_client, realm, prepare_response['state'], prepare_response['nonce'],
                                              _OidcExchange.callback_path)
        return RefreshingEsClient(
            es_client,
            auth_response['username'],
            auth_response['access_token'],
            auth_response['refresh_token'],
            auth_response['expires_in'],
            name=name)
    finally:
        _OidcExchange.callback_path = None


def _oidc_prepare(es_client, realm: str):
    return es_client.perform_request(
        'POST',
        '/_security/oidc/prepare',
        json.dumps({
            'realm': realm,
        }),
        deserialize_it=True
    )


def _oidc_do_authenticate(es_client, realm: str, state: str, nonce: str, redirect_uri: str):
    response = es_client.perform_request(
        'POST',
        '/_security/oidc/authenticate',
        json.dumps({
            'realm': realm,
            'state': state,
            'nonce': nonce,
            'redirect_uri': redirect_uri,
        }),
        deserialize_it=True
    )
    return response


def _oidc_start_http_server(callback_port):
    from peek import __file__ as package_root
    package_root = os.path.dirname(package_root)
    httpd = HTTPServer(('localhost', int(callback_port)), CallbackHTTPRequestHandler)
    keyfile = os.path.join(package_root, 'certs', 'key.pem')
    certfile = os.path.join(package_root, 'certs', 'cert.pem')
    httpd.socket = ssl.wrap_socket(
        httpd.socket,
        keyfile=keyfile,
        certfile=certfile,
        server_side=True)

    t = Thread(target=httpd.serve_forever, daemon=True)
    t.start()
    return httpd


class OidcAuthenticateFunc:
    def __call__(self, app, **options):
        realm = options.get('realm', 'oidc1')
        conn = options.get('conn', None)
        oidc_es_client = oidc_authenticate(
            app.es_client_manager.current if conn is None else app.es_client_manager.get_client(conn),
            realm,
            options.get('callback_port', '5601'),
            name=options.get('name', None),
        )
        app.es_client_manager.add(oidc_es_client)
        return json.dumps({'username': oidc_es_client.username, 'realm': 'realm'})

    @property
    def options(self):
        return {'realm': 'oidc1', 'callback_port': '5601', 'name': None, 'conn': None}

    @property
    def description(self):
        return 'Start OIDC authentication flow'



import json
import logging
import os
import ssl
import time
import urllib
import webbrowser
from http.server import HTTPServer, BaseHTTPRequestHandler
from threading import Thread
from typing import Any, Optional

from peek.connection import EsClient, RefreshingEsClient
from peek.errors import PeekError

_logger = logging.getLogger(__name__)


class _SamlExchange:
    callback_path = None


class CallbackHTTPRequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        pass

    def do_POST(self):
        content_length = int(self.headers['Content-Length'])
        body = self.rfile.read(content_length)
        _SamlExchange.callback_path = body
        self.send_response(200)
        self.end_headers()
        self.wfile.write(b'Callback received, you can now close the browser tab.')

    def log_message(self, fmt: str, *args: Any) -> None:
        _logger.info("%s - - [%s] %s\n" %
                     (self.address_string(),
                      self.log_date_time_string(),
                      fmt % args))


def saml_authenticate(es_client: EsClient, realm: str, callback_port: str, name: Optional[str]):
    _logger.info(f'SAML authenticate for realm {realm!r} and callback port {callback_port!r}')
    prepare_response = _saml_prepare(es_client, realm)
    httpd = _saml_start_http_server(callback_port)
    print('Please use browser to complete authentication against the idP')
    webbrowser.open(prepare_response['redirect'])
    while _SamlExchange.callback_path is None:
        time.sleep(0.1)
    try:
        httpd.shutdown()
        query = urllib.parse.parse_qs(_SamlExchange.callback_path.decode())
        if 'SAMLResponse' not in query or len(query['SAMLResponse']) != 1:
            raise PeekError(f'Invalid saml callback response: {_SamlExchange.callback_path!r}')
        content = query['SAMLResponse'][0]
        auth_response = _saml_do_authenticate(es_client, realm, prepare_response['id'], content)
        return RefreshingEsClient(
            es_client,
            auth_response['username'],
            auth_response['access_token'],
            auth_response['refresh_token'],
            auth_response['expires_in'],
            name=name)
    finally:
        _SamlExchange.callback_path = None


def _saml_prepare(es_client, realm: str):
    return es_client.perform_request(
        'POST',
        '/_security/saml/prepare',
        json.dumps({
            'realm': realm,
        }),
        deserialize_it=True
    )


def _saml_do_authenticate(es_client, realm: str, _id: str, content: str):
    response = es_client.perform_request(
        'POST',
        '/_security/saml/authenticate',
        json.dumps({
            'realm': realm,
            'ids': [_id],
            'content': content,
        }),
        deserialize_it=True
    )
    return response


def _saml_start_http_server(callback_port):
    from peek import __file__ as package_root
    package_root = os.path.dirname(package_root)
    httpd = HTTPServer(('localhost', int(callback_port)), CallbackHTTPRequestHandler)
    keyfile = os.path.join(package_root, 'certs', 'key.pem')
    certfile = os.path.join(package_root, 'certs', 'cert.pem')
    httpd.socket = ssl.wrap_socket(
        httpd.socket,
        keyfile=keyfile,
        certfile=certfile,
        server_side=True)

    t = Thread(target=httpd.serve_forever, daemon=True)
    t.start()
    return httpd


class SamlAuthenticateFunc:
    def __call__(self, app, **options):
        realm = options.get('realm', 'saml1')
        conn = options.get('conn', None)
        saml_es_client = saml_authenticate(
            app.es_client_manager.current if conn is None else app.es_client_manager.get_client(conn),
            realm,
            options.get('callback_port', '5601'),
            name=options.get('name', None),
        )
        app.es_client_manager.add(saml_es_client)
        return json.dumps({'username': saml_es_client.username, 'realm': 'realm'})

    @property
    def options(self):
        return {'realm': 'saml1', 'callback_port': '5601', 'name': None, 'conn': None}

    @property
    def description(self):
        return 'Start SAML authentication flow'



import pytest
from pygments.token import Token, Name, Whitespace

from peek.common import PeekToken
from peek.lexers import PeekLexer, UrlPathLexer, BinOp


@pytest.fixture
def peek_lexer():
    return PeekLexer()


@pytest.fixture
def payload_lexer():
    return PeekLexer(stack=('dict',))


def do_test(lexer, text, error_tokens=None):
    error_tokens = error_tokens or []
    tokens = []
    for t in lexer.get_tokens_unprocessed(text):
        print(t)
        tokens.append(t)
        if t[1] is Token.Error and error_tokens:
            assert t == error_tokens[0]
            error_tokens.pop(0)
            continue
        assert t[1] is not Token.Error, t
    return tokens


def test_numbers(payload_lexer):
    text = """{"numbers": [42, 4.2, 0.42, 42e+1, 4.2e-1, .42, -42, -4.2, -.42]}"""
    for t in payload_lexer.get_tokens_unprocessed(text):
        assert t[1] is not Token.Error, t


def test_es_api_calls(peek_lexer):
    do_test(peek_lexer, """get /some/path
{}

get /another/path a=b c=d
{"foo": "bar"}
{"ok": [42]}

get /yet/another // comment
// this comment is ok
{ // inner comment
  "ok": // trailing comment again
    [ // something
      42, // here
    ],
}
{"some": "other"}
""")


def test_es_api_calls_2(peek_lexer):
    do_test(peek_lexer, """get abc

post abc/_doc
{ "foo":
         "bar"
}

conn foo=bar  // comment
get abc
""")


def test_es_api_call_path_variable(peek_lexer):
    do_test(peek_lexer, '''GET ("hello" + "world" + 1)''')


def test_es_api_call_payloads(peek_lexer):
    do_test(peek_lexer, """get abc
  // comment is ok
 { }
    // another comment is fine
  {}
// yet another one
{ }

get xyz
{}
{}
""")


def test_es_api_call_file(peek_lexer):
    do_test(peek_lexer, """get _abc
@file
""")


def test_func_calls(peek_lexer):
    do_test(peek_lexer, """conn 1
conn "a" foo=1 c=bar

f a b c // this is ok

t 1 2 3 foo=bar // comment

f1 @abc // symbol is fine

g 1 b=[3,4] x={
"a": // inner
 1} // ok""")


def test_continuous_statements(peek_lexer):
    do_test(peek_lexer, """get abc
get xyz
connect 1 2 3
put xyz/_doc
{}
post qwer/_doc
{
  "a": "b",
}""")


def test_invalid_000(peek_lexer):
    do_test(peek_lexer, """conn 1 2 3 foo=bar /""", error_tokens=[(19, Token.Error, '/')])
    do_test(peek_lexer, """get / /""", error_tokens=[(6, Token.Error, '/')])


def test_invalid_001(peek_lexer):
    do_test(peek_lexer, """get /abc

{}
""", error_tokens=[PeekToken(index=10, ttype=Token.Error, value='{'),
                   PeekToken(index=11, ttype=Token.Error, value='}')])


def test_es_api_and_func_calls(peek_lexer):
    do_test(peek_lexer, """ // begining comment
get / conn=1 // first api call

connection 1 // set conneciton to 1
get / // this is the same as the first api call

f a b c q=42

put /
{}

g 42""")


def test(peek_lexer):
    do_test(peek_lexer, """ // some comments to start the day
get /some/path with=1.2 option=foo another="bar" // trailing comment
{ "hello": ["world", 1, '1'], }

connect 1 foo="bar" ok=good

put /here
{ 'some more': "things to do" }
{ 'even more': { "nest": "here", } }
""")


def test_minimal(peek_lexer):
    do_test(peek_lexer, text="""c 1 a=b""")


def test_comment(peek_lexer):
    do_test(peek_lexer, text="""// comment
""")


def test_shellout(peek_lexer):
    do_test(peek_lexer, text="""!ls
""")


def test_func_call_connect(peek_lexer):
    do_test(peek_lexer, text="""connect hosts='https://localhost:9200' username='foo'""")


def test_builtin_as_kv_value(peek_lexer):
    tokens = do_test(peek_lexer, text="""connect use_ssl=true""")
    assert tokens[-1][1] is Name.Builtin


def test_dot_notation(peek_lexer):
    tokens = [t for t in do_test(peek_lexer, text='''f a.1.2.5''') if (t.ttype is BinOp and t.value == '.')]
    assert 3 == len(tokens)


def test_expressions(peek_lexer):
    tokens = do_test(peek_lexer, text='''f kw=1 * (2 + (3 - foo)) / a.4.2."a key".(a + b.3).5."c" % 5.6 2+3/(a)
''')
    print(tokens)


def test_expressions_data_structure(peek_lexer):
    tokens = do_test(peek_lexer, text='''GET /
{
  "a": "foo" + a.3."ab".(foo*3),
  "b": 1 + 3 * (2 - 1)
}''')
    print(tokens)


def test_dict_keys(peek_lexer):
    tokens = do_test(peek_lexer, text='''f {
    "a": 4.2,
    "a" + b : "c",
    """a""" + c: 5,
}''')
    print(tokens)


def test_operator_cannot_on_separate_line(peek_lexer):
    with pytest.raises(AssertionError) as e:
        do_test(peek_lexer, text='''f 1
+ 2''')

    assert "value='+'" in e.value.args[0]


def test_expression_pos_neg(peek_lexer):
    tokens = do_test(peek_lexer, text='''f -1 + -a + -(3 - -2)''')
    print(tokens)


def test_function_expr(peek_lexer):
    tokens = do_test(peek_lexer, text='''f 3 -f(a b=1 c="ok") "hello" 4.2 {"a": f(9 c=3) }''')
    print(tokens)


def test_symbol(peek_lexer):
    tokens = do_test(peek_lexer, text='''f a.@b.c.@d.1''')
    assert len([t for t in tokens if (t.ttype is BinOp and t.value == '.')]) == 4


def test_function_expr_chain(peek_lexer):
    do_test(peek_lexer, text='''f a.b.c().d.e().f.1.2 {}()''')


def test_for_stmt(peek_lexer):
    do_test(peek_lexer, text='''for i in [1,2,3] {
    do_something a b c d=1
    !shellout
    GET /api/path
    {}
    for j in [4,5,6] {
        nested_call
        GET /
        {}
        echo hello
    }
    another_call
    history
}''')


def test_whitespace_comma(peek_lexer):
    tokens = do_test(peek_lexer, text='''f a,b,c''')
    tokens = [t for t in tokens if t.ttype is not Whitespace]
    assert len(tokens) == 4

    tokens = do_test(peek_lexer, text='''f g(a,b,c)''')
    tokens = [t for t in tokens if t.ttype is not Whitespace]
    assert len(tokens) == 7


@pytest.fixture
def url_path_lexer():
    return UrlPathLexer()


def test_url_path_empty(url_path_lexer):
    do_test(url_path_lexer, '')


def test_url_path_only(url_path_lexer):
    do_test(url_path_lexer, '/a/b/c')
    do_test(url_path_lexer, 'a/b/c/')
    do_test(url_path_lexer, '/a/b/c?')


def test_url_with_query(url_path_lexer):
    do_test(url_path_lexer, '/a/b/c?foo=bar&hello=42')
    do_test(url_path_lexer, '/a/b/c?foo=bar&name&pretty=')


def test_invalid_url(url_path_lexer):
    do_test(url_path_lexer, '/?=', error_tokens=[
        PeekToken(index=2, ttype=Token.Error, value='=')
    ])



from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from typing import List, Union

from peek.common import PeekToken


class Visitor(metaclass=ABCMeta):

    def __init__(self):
        self._consumers = []

    def visit_es_api_call_node(self, node):
        raise NotImplementedError()

    def visit_func_call_node(self, node):
        raise NotImplementedError()

    def visit_let_node(self, node):
        raise NotImplementedError()

    def visit_shell_out_node(self, node):
        raise NotImplementedError()

    def visit_for_in_node(self, node):
        raise NotImplementedError()

    def visit_name_node(self, node):
        raise NotImplementedError()

    def visit_symbol_node(self, node):
        raise NotImplementedError()

    def visit_text_node(self, node):
        raise NotImplementedError()

    def visit_key_value_node(self, node):
        raise NotImplementedError()

    def visit_dict_node(self, node):
        raise NotImplementedError()

    def visit_array_node(self, node):
        raise NotImplementedError()

    def visit_string_node(self, node):
        raise NotImplementedError()

    def visit_number_node(self, node):
        raise NotImplementedError()

    def visit_bin_op_node(self, node):
        raise NotImplementedError()

    def visit_unary_op_node(self, node):
        raise NotImplementedError()

    def visit_group_node(self, node):
        raise NotImplementedError()

    def push_consumer(self, consumer):
        self._consumers.append(consumer)

    def pop_consumer(self):
        if not self._consumers:
            raise IndexError('No consumer')
        self._consumers = self._consumers[:-1]

    def consume(self, *args, **kwargs):
        if not self._consumers:
            raise IndexError('No consumer')
        self._consumers[-1](*args, **kwargs)

    @contextmanager
    def consumer(self, c):
        self.push_consumer(c)
        try:
            yield
        finally:
            self.pop_consumer()


class Node(metaclass=ABCMeta):

    @abstractmethod
    def accept(self, visitor: Visitor):
        pass

    @abstractmethod
    def tokens(self):
        pass

    def __str__(self):
        return self.tokens()[0].value

    def __repr__(self):
        return repr(str(self))


class NameNode(Node):

    def __init__(self, token: PeekToken):
        self.token = token

    def accept(self, visitor):
        visitor.visit_name_node(self)

    def tokens(self):
        return [self.token]


class TextNode(Node):

    def __init__(self, token: PeekToken):
        self.token = token

    def accept(self, visitor):
        visitor.visit_text_node(self)

    def tokens(self):
        return [self.token]


class SymbolNode(Node):

    def __init__(self, token: PeekToken):
        self.token = token

    def accept(self, visitor):
        visitor.visit_symbol_node(self)

    def tokens(self):
        return [self.token]


class KeyValueNode(Node):

    def __init__(self, key_node: Node, value_node: Node):
        self.key_node = key_node
        self.value_node = value_node

    def accept(self, visitor):
        visitor.visit_key_value_node(self)

    def tokens(self):
        return self.key_node.tokens() + self.value_node.tokens()

    def __str__(self):
        return f'{self.key_node}:{self.value_node}'


class DictNode(Node):

    def __init__(self, kv_nodes: List[KeyValueNode]):
        self.kv_nodes = kv_nodes

    def accept(self, visitor: Visitor):
        visitor.visit_dict_node(self)

    def tokens(self):
        tokens = []
        for n in self.kv_nodes:
            tokens += n.tokens()
        return tokens

    def __str__(self):
        parts = ['{']
        for i, n in enumerate(self.kv_nodes):
            parts.append(str(n))
            if i < len(self.kv_nodes) - 1:
                parts.append(',')
        parts.append('}')
        return ''.join(parts)


class ArrayNode(Node):

    def __init__(self, value_nodes: List[Node]):
        self.value_nodes = value_nodes

    def accept(self, visitor: Visitor):
        return visitor.visit_array_node(self)

    def tokens(self):
        tokens = []
        for n in self.value_nodes:
            tokens += n.tokens()
        return tokens

    def __str__(self):
        parts = ['[']
        for i, n in enumerate(self.value_nodes):
            parts.append(str(n))
            if i < len(self.value_nodes) - 1:
                parts.append(',')
        parts.append(']')
        return ''.join(parts)


class StringNode(Node):

    def __init__(self, token: PeekToken):
        self.token = token

    def accept(self, visitor: Visitor):
        visitor.visit_string_node(self)

    def tokens(self):
        return [self.token]


class NumberNode(Node):

    def __init__(self, token: PeekToken):
        self.token = token

    def accept(self, visitor: Visitor):
        visitor.visit_number_node(self)

    def tokens(self):
        return [self.token]


class UnaryOpNode(Node):

    def __init__(self, op_token: PeekToken, operand_node: Node):
        self.op_token = op_token
        self.operand_node = operand_node

    def accept(self, visitor: Visitor):
        visitor.visit_unary_op_node(self)

    def tokens(self):
        tokens = [self.op_token]
        tokens += self.operand_node.tokens()
        return tokens


class BinOpNode(Node):

    def __init__(self, op_token: PeekToken, left_node: Node, right_node: Node):
        self.op_token = op_token
        self.left_node = left_node
        self.right_node = right_node

    def accept(self, visitor: Visitor):
        visitor.visit_bin_op_node(self)

    def tokens(self):
        tokens = self.left_node.tokens()
        tokens.append(self.op_token)
        tokens += self.right_node.tokens()
        return tokens

    def __str__(self):
        return f'{self.left_node} {self.op_token.value} {self.right_node}'


class GroupNode(Node):
    """
    A decorator node that indicates there are parenthesis around the node
    """

    def __init__(self, node: Node, paren_left: PeekToken, paren_right: PeekToken):
        self.grouped = node
        self.paren_left = paren_left
        self.paren_right = paren_right

    def accept(self, visitor: Visitor):
        visitor.visit_group_node(self)

    def tokens(self):
        tokens = [self.paren_left]
        tokens += self.grouped.tokens()
        tokens.append(self.paren_right)
        return tokens

    def __str__(self):
        return f'({self.grouped})'


class EsApiCallNode(Node, metaclass=ABCMeta):

    def __init__(self, method_node: NameNode, path_node: Union[TextNode, GroupNode], options_node: DictNode):
        self.method_node = method_node
        self.path_node = path_node
        self.options_node = options_node

    def accept(self, visitor):
        visitor.visit_es_api_call_node(self)

    def tokens(self):
        tokens = self.method_node.tokens() + self.path_node.tokens()
        tokens += self.options_node.tokens()
        return tokens

    @property
    def method(self):
        return self.method_node.token.value.upper()

    @property
    def path(self):
        if isinstance(self.path_node, TextNode):
            path = self.path_node.token.value
            return path if path.startswith('/') else ('/' + path)
        else:
            raise ValueError('simple path is only available when path node is simple text')


class EsApiCallInlinePayloadNode(EsApiCallNode):

    def __init__(self, method_node: NameNode, path_node: Union[TextNode, GroupNode], options_node: DictNode,
                 dict_nodes: List[DictNode]):
        super().__init__(method_node, path_node, options_node)
        self.dict_nodes = dict_nodes

    def tokens(self):
        tokens = super().tokens()
        for n in self.dict_nodes:
            tokens += n.tokens
        return tokens

    def __str__(self):
        parts = [str(self.method_node), ' ', str(self.path_node), ' ', str(self.options_node), '\n']
        for n in self.dict_nodes:
            parts.append(str(n))
            parts.append('\n')
        return ''.join(parts)


class EsApiCallFilePayloadNode(EsApiCallNode):

    def __init__(self, method_node: NameNode, path_node: Union[TextNode, GroupNode], options_node: DictNode,
                 file_node: TextNode):
        super().__init__(method_node, path_node, options_node)
        self.file_node = file_node

    def tokens(self):
        tokens = super().tokens()
        tokens.append(self.file_node.token)
        return tokens

    def __str__(self):
        parts = [str(self.method_node), ' ', str(self.path_node), ' ', str(self.options_node), '\n',
                 '@', str(self.file_node), '\n']
        return ''.join(parts)


class FuncCallNode(Node):

    def __init__(self, name_node: Node, symbols_node: ArrayNode, args_node: ArrayNode,
                 kwargs_node: DictNode, is_stmt=True):
        self.name_node = name_node
        self.symbols_node = symbols_node
        self.args_node = args_node
        self.kwargs_node = kwargs_node
        self.is_stmt = is_stmt

    def accept(self, visitor: Visitor):
        visitor.visit_func_call_node(self)

    def tokens(self):
        tokens = self.name_node.tokens()
        tokens += self.symbols_node.tokens()
        tokens += self.args_node.tokens()
        tokens += self.kwargs_node.tokens()
        return tokens

    def __str__(self):
        parts = [str(self.name_node), ' ',
                 str(self.symbols_node), ' ',
                 str(self.args_node), ' ',
                 str(self.kwargs_node), '\n' if self.is_stmt else '']
        return ''.join(parts)


class LetNode(Node):

    def __init__(self, assignments_node: DictNode):
        self.assignments_node = assignments_node

    def accept(self, visitor: Visitor):
        visitor.visit_let_node(self)

    def tokens(self):
        return self.assignments_node.tokens()

    def __str__(self):
        parts = ['let', ' ',
                 str(self.assignments_node), '\n']
        return ''.join(parts)


class ForInNode(Node):

    def __init__(self, item: NameNode, items: Node, suite: List[Node]):
        self.item = item
        self.items = items
        self.suite = suite

    def accept(self, visitor: Visitor):
        visitor.visit_for_in_node(self)

    def tokens(self):
        tokens = self.item.tokens() + self.items.tokens()
        for node in self.suite:
            tokens += node.tokens()
        return tokens

    def __str__(self):
        parts = [
            'for ',
            self.item.token.value,
            ' in ',
            str(self.items),
            '{\n'
        ]
        for node in self.suite:
            parts.append(str(node))
        parts.append('}\n')
        return ''.join(parts)


class ShellOutNode(Node):

    def __init__(self, text_node: TextNode):
        self.text_node = text_node

    def accept(self, visitor: Visitor):
        visitor.visit_shell_out_node(self)

    @property
    def command(self):
        return self.text_node.token.value

    def tokens(self):
        return self.text_node.tokens()

    def __str__(self):
        parts = ['!', ' ',
                 str(self.text_node), '\n']
        return ''.join(parts)



"""Main module."""
import logging
import logging.handlers
import sys
from datetime import datetime
from typing import Iterable

from prompt_toolkit import PromptSession, prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.layout.processors import HighlightMatchingBracketProcessor
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import style_from_pygments_cls

from peek.capture import NoOpCapture, FileCapture
from peek.common import NONE_NS
from peek.completer import PeekCompleter
from peek.config import get_config, config_location
from peek.connection import EsClientManager
from peek.display import Display
from peek.errors import PeekError, PeekSyntaxError
from peek.history import SqLiteHistory
from peek.key_bindings import key_bindings
from peek.lexers import PeekLexer, PeekStyle, Heading, TipsMinor
from peek.natives import ConnectFunc
from peek.parser import PeekParser
from peek.vm import PeekVM

_logger = logging.getLogger(__name__)


class PeekApp:

    def __init__(self,
                 batch_mode=False,
                 config_file: str = None,
                 extra_config_options: Iterable[str] = None,
                 cli_ns=NONE_NS):
        self._should_exit = False
        self._preserved_text = ''
        self.capture = NoOpCapture()
        self.batch_mode = batch_mode
        self.config = get_config(config_file, extra_config_options)
        self.cli_ns = cli_ns
        self._init_logging()
        self.es_client_manager = EsClientManager()
        self._init_es_client()
        self.history = SqLiteHistory(self.config.as_int('history_max'))
        self.prompt = self._init_prompt()
        self.display = Display(self)
        self.parser = PeekParser()
        self.vm = self._init_vm()

        # TODO: better name for signal payload json reformat
        self.is_pretty = True

    def run(self):
        while True:
            try:
                text: str = self.prompt.prompt(
                    message=self._get_message(),
                    default=self._get_default_text(),
                )
                _logger.debug(f'input: {text!r}')
                if self._should_exit:
                    raise EOFError()
                if text.strip() == '':
                    continue
                self.process_input(text)
            except KeyboardInterrupt:
                continue
            except EOFError:
                break

    def process_input(self, text, echo=False):
        try:
            nodes = self.parser.parse(text)
        except PeekSyntaxError as e:
            self.display.error(e)
            return

        for node in nodes:
            try:
                if echo:
                    self.display.info(str(node))
                self.execute_node(node)
            except PeekError as e:
                self.display.error(e)
            except Exception as e:
                self.display.error(e)
                _logger.exception('Error on node execution')

    def execute_node(self, node):
        self.vm.execute_node(node)

    def signal_exit(self):
        self._should_exit = True

    def reset(self):
        self._init_es_client()
        for _ in range(len(self.es_client_manager.clients()) - 1):
            self.es_client_manager.remove_client(0)
        self.vm.context = {}

    def start_capture(self, f=None):
        if not isinstance(self.capture, NoOpCapture):
            raise PeekError(f'Cannot capture when one is currently running: {self.capture.status()}')

        if f is None:
            f = f'{datetime.now().strftime("%Y%m%d%H%M%S")}.es'
        self.capture = FileCapture(f)
        return self.capture.status()

    def stop_capture(self):
        if not isinstance(self.capture, NoOpCapture):
            self.capture.stop()
            self.capture = NoOpCapture()
            return 'Capture stopped'
        else:
            return 'No capture is running'

    @property
    def preserved_text(self):
        return self._preserved_text

    @preserved_text.setter
    def preserved_text(self, value: str):
        self._preserved_text = value

    def input(self, message='', is_secret=False):
        return prompt(message=message, is_password=is_secret)

    def _get_message(self):
        idx = self.es_client_manager.clients().index(self.es_client_manager.current)
        info_line = f' [{idx}] {self.es_client_manager.current}'
        if len(info_line) > 100:
            info_line = info_line[:97] + '...'
        return FormattedText([
            (PeekStyle.styles[Heading], '>>>'),
            (PeekStyle.styles[TipsMinor], info_line + '\n'),
        ])

    def _get_default_text(self):
        text = self.preserved_text
        if text:
            self.preserved_text = ''
        return text

    def _init_logging(self):
        log_file = self.config['log_file']
        log_level = self.config['log_level'].upper()
        if log_level == 'NONE':
            handler = logging.NullHandler()
        elif log_file == 'stderr':
            handler = logging.StreamHandler(sys.stderr)
        elif log_file == 'stdout':
            handler = logging.StreamHandler(sys.stdout)
        else:
            handler = logging.handlers.RotatingFileHandler(
                config_location() + log_file,
                maxBytes=10 * 1024 * 1024, backupCount=5)

        log_level = getattr(logging, log_level, logging.WARNING)

        formatter = logging.Formatter(
            '%(asctime)s (%(process)d/%(threadName)s) '
            '%(name)s %(levelname)s - %(message)s'
        )
        handler.setFormatter(formatter)

        root_logger = logging.getLogger(__package__)
        root_logger.addHandler(handler)
        root_logger.setLevel(log_level)

    def _init_prompt(self):
        if self.batch_mode:
            return None
        else:
            try:
                from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
                clipboard = PyperclipClipboard()
            except ImportError:
                clipboard = None

            return PromptSession(
                message=self._get_message(),
                style=style_from_pygments_cls(PeekStyle),
                lexer=PygmentsLexer(PeekLexer),
                auto_suggest=AutoSuggestFromHistory(),
                completer=PeekCompleter(self),
                history=self.history,
                multiline=True,
                key_bindings=key_bindings(self),
                enable_open_in_editor=True,
                enable_system_prompt=False,
                enable_suspend=True,
                search_ignore_case=True,
                clipboard=clipboard,
                mouse_support=False,
                swap_light_and_dark_colors=self.config.as_bool('swap_colour'),
                input_processors=[
                    HighlightMatchingBracketProcessor(chars="[](){}"),
                ],
            )

    def _init_es_client(self):
        options = {}
        keys = [
            'name',
            'hosts',
            'cloud_id',
            'username',
            'password',
            'api_key',
            'token',
            'use_ssl',
            'verify_certs',
            'assert_hostname',
            'ca_certs',
            'client_cert',
            'client_key',
            'force_prompt',
            'no_prompt',
        ]
        for k in keys:
            v = getattr(self.cli_ns, k, None)
            if v is not None:
                options[k] = v

        ConnectFunc()(self, **options)

    def _init_vm(self):
        return PeekVM(self)



import ast
import itertools
import json
import logging
import os
from typing import Iterable, List, Dict, Tuple, Optional

from prompt_toolkit.completion import Completer, CompleteEvent, Completion, WordCompleter, FuzzyCompleter, PathCompleter
from prompt_toolkit.contrib.completers import SystemCompleter
from prompt_toolkit.document import Document
from pygments.token import Error, Name, Literal, String

from peek.common import PeekToken
from peek.errors import PeekError
from peek.lexers import PeekLexer, UrlPathLexer, PathPart, ParamName, Ampersand, QuestionMark, Slash, HttpMethod, \
    FuncName, OptionName, Assign, CurlyLeft, CurlyRight, DictKey, ShellOut, At
from peek.parser import process_tokens

_logger = logging.getLogger(__name__)

_HTTP_METHOD_COMPLETER = WordCompleter(['GET', 'POST', 'PUT', 'DELETE'], ignore_case=True)

_ES_API_CALL_OPTION_COMPLETER = WordCompleter([w + '=' for w in sorted(['conn', 'runas', 'headers', 'xoid'])])

_PATH_COMPLETER = PathCompleter(expanduser=True)
_SYSTEM_COMPLETER = SystemCompleter()


class PeekCompleter(Completer):

    def __init__(self, app):
        self.app = app
        self.lexer = PeekLexer()
        self.url_path_lexer = UrlPathLexer()
        from peek import __file__ as package_root
        package_root = os.path.dirname(package_root)
        kibana_dir = app.config['kibana_dir'] or os.path.join(package_root, 'specs', 'kibana-7.8.1')
        if self.app.config.as_bool('load_api_specs'):
            self.specs = load_specs(kibana_dir)
        else:
            self.specs = {}

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        _logger.debug(f'doc: {document}, event: {complete_event}')

        text = document.text_before_cursor
        _logger.debug(f'text before cursor: {text!r}')

        # Parse tokens for only the text before cursor and merge certain consecutive tokens
        tokens = process_tokens(self.lexer.get_tokens_unprocessed(text))
        _logger.debug(f'processed tokens: {tokens}')

        # Nothing to complete if no significant tokens are found
        if len(tokens) == 0:
            return []

        idx_head_token, head_token = find_beginning_token(tokens)
        _logger.debug(f'head token: {head_token} at {idx_head_token}')
        if head_token is None:
            return []

        pos_head_token = document.translate_index_to_position(head_token.index)
        last_token = tokens[-1]
        pos_cursor = document.translate_index_to_position(document.cursor_position)

        # Cursor is on a non-white token
        is_cursor_on_non_white_token = (last_token.index < document.cursor_position
                                        <= (last_token.index + len(last_token.value)))

        if head_token.ttype is ShellOut:
            _logger.debug('Completing for shell out')
            return _SYSTEM_COMPLETER.get_completions(Document(text[head_token.index + 1:]), complete_event)

        if is_cursor_on_non_white_token:
            _logger.debug(f'Cursor token: {last_token}')
            if last_token.ttype in (HttpMethod, FuncName):
                _logger.debug(f'Completing function/http method name: {last_token}')
                return itertools.chain(
                    _HTTP_METHOD_COMPLETER.get_completions(document, complete_event),
                    WordCompleter(['for', 'let']).get_completions(document, complete_event),
                    WordCompleter(self.app.vm.functions.keys()).get_completions(document, complete_event)
                )

            # The token right before cursor is HttpMethod, go for path completion
            if head_token.ttype is HttpMethod and idx_head_token == len(tokens) - 2 and last_token.ttype is Literal:
                return self._complete_http_path(document, complete_event, tokens[idx_head_token:])

            if head_token.value == 'run' and idx_head_token == len(tokens) - 2 and last_token.ttype in String:
                arg = text.split()[-1]
                if last_token.ttype in (String.Single, String.Double):
                    doc = Document(arg[1:])
                else:
                    doc = Document(arg[3:])
                _logger.debug(f'Completing for file path for run: {doc}')
                return _PATH_COMPLETER.get_completions(doc, complete_event)

            if head_token.ttype is HttpMethod:
                if last_token.ttype is At:
                    return _PATH_COMPLETER.get_completions(Document(text[last_token.index + 1:]), complete_event)
                elif last_token.ttype is Literal and len(tokens) > 1 and tokens[-2].ttype is At:
                    return _PATH_COMPLETER.get_completions(Document(last_token.value), complete_event)
                elif last_token.ttype is DictKey:
                    return self._complete_payload(document, complete_event, tokens[idx_head_token:])
            elif head_token.ttype is FuncName:
                if last_token.ttype in (At, Literal):
                    return self._complete_options(document, complete_event, tokens[idx_head_token:],
                                                  is_cursor_on_non_white_token)

            if last_token.ttype in (OptionName, Name):
                return self._complete_options(document, complete_event, tokens[idx_head_token:],
                                              is_cursor_on_non_white_token)

            return []

        else:  # Cursor is on a whitespace
            _logger.debug('cursor is after the last token')
            # Cursor is on whitespace after the last non-white token
            if pos_head_token[0] != pos_cursor[0] or last_token.ttype is HttpMethod:
                # Cursor is on separate line or immediately after http method
                return []
            else:
                # Cursor is on the same line and at the end of an ES API or func call
                _logger.debug('cursor is at the end of a statement')
                return self._complete_options(document, complete_event, tokens[idx_head_token:],
                                              is_cursor_on_non_white_token)

    def _complete_http_path(self, document: Document, complete_event: CompleteEvent, tokens) -> Iterable[Completion]:
        method_token, path_token = tokens[-2], tokens[-1]
        method = method_token.value.upper()
        cursor_position = document.cursor_position - path_token.index
        path = path_token.value[:cursor_position]
        _logger.debug(f'Completing http path: {path!r}')
        path_tokens = list(self.url_path_lexer.get_tokens_unprocessed(path))
        _logger.debug(f'path_tokens: {path_tokens}')
        if not path_tokens:  # empty, should not happen
            return []

        cursor_token = path_tokens[-1]
        _logger.debug(f'cursor_token: {cursor_token}')
        if cursor_token.ttype is Error:
            return []
        elif cursor_token.ttype in (PathPart, Slash):
            return self._complete_http_path_part(document, complete_event, method, path_tokens)
        elif cursor_token.ttype in (ParamName, QuestionMark, Ampersand):
            return self._complete_query_param_name(document, complete_event, method, path_tokens)
        else:  # skip for param value
            return self._complete_query_param_value(document, complete_event, method, path_tokens)

    def _complete_http_path_part(self, document: Document, complete_event: CompleteEvent, method, path_tokens):
        cursor_token = path_tokens[-1]
        _logger.debug(f'Completing path part: {cursor_token}')
        ts = [t.value for t in path_tokens if t.ttype is not Slash]
        if cursor_token.ttype is PathPart:
            ts.pop()
        _logger.debug(f'ts: {ts}')
        candidates = []
        for api_name, api_spec in self.specs.items():
            if method not in api_spec['methods']:
                continue
            for api_path in api_spec['patterns']:
                ps = [p for p in api_path.split('/') if p]
                # Nothing to complete if the candidate is shorter than current input
                if len(ts) >= len(ps):
                    continue
                if not can_match(ts, ps):
                    continue
                candidate = '/'.join(ps[len(ts):])
                candidates.append(Completion(candidate, start_position=0))

        return FuzzyCompleter(ConstantCompleter(candidates)).get_completions(document, complete_event)

    def _complete_query_param_name(self, document: Document, complete_event: CompleteEvent, method, path_tokens):
        _logger.debug(f'Completing query param name: {path_tokens[-1]}')
        ts = [t.value for t in path_tokens if t.ttype is PathPart]
        candidates = set()
        for api_spec in matchable_specs(method, ts, self.specs):
            candidates.update(api_spec['url_params'].keys())

        return FuzzyCompleter(ConstantCompleter(
            [Completion(c, start_position=0) for c in candidates])).get_completions(document, complete_event)

    def _complete_query_param_value(self, document: Document, complete_event: CompleteEvent, method, path_tokens):
        _logger.debug(f'Completing query param value: {path_tokens[-1]}')
        param_name_token = path_tokens[-2] if path_tokens[-1].ttype is Assign else path_tokens[-3]
        _logger.debug(f'Param name token: {param_name_token}')
        ts = [t.value for t in path_tokens if t.ttype is PathPart]
        candidates = set()
        for api_spec in matchable_specs(method, ts, self.specs):
            v = api_spec['url_params'].get(param_name_token.value, None)
            if v is None:
                continue
            if v == '__flag__':
                candidates.update(('true', 'false'))
            elif isinstance(v, list) and len(v) > 0:
                candidates.update(v)

        return FuzzyCompleter(ConstantCompleter(
            [Completion(c, start_position=0) for c in candidates])).get_completions(document, complete_event)

    def _complete_options(self, document: Document, complete_event: CompleteEvent, tokens: List[PeekToken],
                          is_cursor_on_non_white_token):
        _logger.debug(f'Completing for options: {is_cursor_on_non_white_token}')
        head_token, last_token = tokens[0], tokens[-1]

        def _get_option_name_for_value_completion():
            if not is_cursor_on_non_white_token and last_token.ttype is Assign:
                return tokens[-2].value
            elif is_cursor_on_non_white_token and (len(tokens) > 1 and tokens[-2].ttype is Assign):
                return tokens[-3].value
            else:
                return None

        # TODO: For future handle the KeyName or Name is inside a function call, e.g. f(name=..)
        if head_token.ttype is HttpMethod:
            option_name = _get_option_name_for_value_completion()
            if option_name is not None:
                return []  # TODO: complete for value
            else:
                return _ES_API_CALL_OPTION_COMPLETER.get_completions(document, complete_event)
        elif head_token.ttype is FuncName:
            func = self.app.vm.functions.get(head_token.value)
            if func is None or not getattr(func, 'options', None):
                return []
            option_name = _get_option_name_for_value_completion()
            if option_name is not None:
                return []  # TODO: complete for value
            else:
                options = sorted([n if n.startswith('@') else (n + '=') for n in func.options.keys()])
                return WordCompleter(options, WORD=True).get_completions(
                    document, complete_event)
        else:
            return []

    def _complete_payload(self, document: Document, complete_event: CompleteEvent,
                          tokens: List[PeekToken]) -> Iterable[Completion]:
        _logger.debug(f'Completing for payload with tokens: {tokens}')
        method_token, path_token = tokens[0], tokens[1]
        path_tokens = list(self.url_path_lexer.get_tokens_unprocessed(path_token.value))
        ts = [t.value for t in path_tokens if t.ttype is not Slash]

        try:
            api_spec = next(matchable_specs(method_token.value.upper(), ts, self.specs,
                                            required_field='data_autocomplete_rules'))
        except StopIteration:
            return []

        rules = _maybe_unwrap_for_dict(api_spec.get('data_autocomplete_rules', None))
        if rules is None:
            return []

        _logger.debug(f'Found rules from spec: {rules}')

        payload_keys = []
        curly_level = 0
        for t in tokens[2:-1]:
            if t.ttype is CurlyLeft:
                curly_level += 1
            elif t.ttype is CurlyRight:
                curly_level -= 1
                payload_keys.pop()
            elif t.ttype is DictKey:
                if len(payload_keys) == curly_level - 1:
                    payload_keys.append(ast.literal_eval(t.value))
                elif len(payload_keys) == curly_level:
                    payload_keys[-1] = ast.literal_eval(t.value)
                else:
                    raise PeekError(f'Error when counting curly level {curly_level} and keys {payload_keys}')

        _logger.debug(f'Payload status: level: {curly_level}, keys: {payload_keys}')
        if curly_level == 0:  # not even in the first curly bracket, no completion
            return []

        # Remove the payload key that is at the same level
        if curly_level == len(payload_keys):
            payload_keys.pop()

        for k in payload_keys:
            rules = _maybe_unwrap_for_dict(rules.get(k, None))
            if rules is None:
                _logger.debug(f'Rules not available for key: {k}')
                return []

        return FuzzyCompleter(ConstantCompleter(
            [Completion(c, start_position=0) for c in rules.keys()])).get_completions(document, complete_event)


def _maybe_unwrap_for_dict(rules):
    if isinstance(rules, dict):
        return rules
    elif isinstance(rules, list) and len(rules) > 0:
        if isinstance(rules[0], dict):
            return rules[0]
        else:
            return None
    else:
        return None


def find_beginning_token(tokens) -> Tuple[Optional[int], Optional[PeekToken]]:
    for i, t in zip(reversed(range(len(tokens))), tokens[::-1]):
        if t.ttype in (HttpMethod, FuncName, ShellOut):
            return i, t
    return None, None


def matchable_specs(method: str, ts: List[str], specs: Dict, required_field='url_params') -> Dict:
    """
    Find the matchable spec for the given HTTP method and input path.
    """
    for api_name, api_spec in specs.items():
        if method not in api_spec['methods']:
            continue
        if not api_spec.get(required_field, None):
            continue
        matched = False
        for pattern in api_spec['patterns']:
            ps = [p for p in pattern.split('/') if p]
            if len(ts) != len(ps):
                continue
            if not can_match(ts, ps):
                continue
            matched = True
            break
        if matched:
            yield api_spec


def can_match(ts, ps):
    """
    Test whether the input path (ts) can match the candidate path (ps).
    The rule is basically a placeholder can match any string other than
    the ones leading with underscore.
    """
    for t, p in zip(ts, ps):
        if t != p:
            if t.startswith('_'):
                return False
            if not (p.startswith('{') and p.endswith('}')):
                return False
    return True


class ConstantCompleter(Completer):

    def __init__(self, candidates):
        self.candidates = candidates

    def get_completions(self, document: Document, complete_event: CompleteEvent) -> Iterable[Completion]:
        return self.candidates


def load_specs(kibana_dir):
    oss_path = os.path.join(
        kibana_dir, 'src', 'plugins', 'console', 'server', 'lib', 'spec_definitions')
    xpack_path = os.path.join(
        kibana_dir, 'x-pack', 'plugins', 'console_extensions', 'server', 'lib', 'spec_definitions')
    specs = _load_json_specs(os.path.join(oss_path, 'json'))
    specs.update(_load_json_specs(os.path.join(xpack_path, 'json')))
    return specs


def _load_json_specs(base_dir):
    _logger.debug(f'Loading json specs from: {base_dir!r}')
    specs = {}
    for sub_dir in ('generated', 'overrides'):
        d = os.path.join(base_dir, sub_dir)
        if not os.path.exists(d):
            _logger.warning(f'JSON specs directory does not exist: {d}')
            continue
        for f in os.listdir(d):
            if f == '_common.json':
                continue
            with open(os.path.join(d, f)) as ins:
                spec = json.load(ins)
            if sub_dir == 'generated':
                specs.update(spec)
            else:
                for k, v in spec.items():
                    if k in specs:
                        specs[k].update(v)
                    else:
                        if k.startswith('xpack.'):
                            specs[k[6:]].update(v)
                        else:
                            specs['xpack.' + k].update(v)

    return specs



import pytest
from pygments.token import String, Whitespace, Comment

from peek.ast import EsApiCallNode, ShellOutNode
from peek.errors import PeekSyntaxError
from peek.lexers import CurlyLeft, CurlyRight
from peek.parser import PeekParser, process_tokens, PeekToken


@pytest.fixture
def parser():
    return PeekParser()


def test_process_tokens():
    tokens = [PeekToken(*t) for t in (
        (0, String.Double, '"'), (1, String.Double, 'str'), (4, String.Double, '"'),
        (5, CurlyLeft, '{'), (6, CurlyLeft, '{'), (7, Whitespace, '  '),
        (10, CurlyRight, '}'), (11, Whitespace, ' '), (12, CurlyRight, '}'),
        (13, String.Double, '"'), (14, String.Double, 'd'), (15, String.Double, '"'),
        (16, String.Single, "'"), (17, String.Single, 's'), (18, String.Single, "'"),
        (19, Comment.Single, "# comment")
    )]
    processed_tokens = process_tokens(tokens)
    assert processed_tokens == [
        (0, String.Double, '"str"'),
        (5, CurlyLeft, '{'), (6, CurlyLeft, '{'),
        (10, CurlyRight, '}'), (12, CurlyRight, '}'),
        (13, String.Double, '"d"'),
        (16, String.Single, "'s'"),
    ]


def test_parser_comment(parser):
    text = """// comment
"""
    nodes = parser.parse(text)
    assert len(nodes) == 0


def test_parser_shellout(parser):
    text = """!ls something another
"""
    nodes = parser.parse(text)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, ShellOutNode)
    assert n.command == 'ls something another'


def test_parser_single_es_api_call(parser):
    text = """get /abc"""
    nodes = parser.parse(text)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, EsApiCallNode)
    assert n.method == 'GET'
    assert n.path == '/abc'
    assert len(n.dict_nodes) == 0


def test_api_path_variable(parser):
    text = """get ("a" + "bc")"""
    nodes = parser.parse(text)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, EsApiCallNode)
    assert n.method == 'GET'
    assert str(n.path_node) == '("a" + "bc")'


def test_parser_multiple_simple_statements(parser):
    text = """get abc

post abc/_doc
{ "foo":
         "bar"
}

conn foo=bar  // comment
get abc
post xyz/_doc
{"index": "asfa"}
// comment
{"again": [{"ok": 1}]}
get xyz/_doc/1 // comment
session @info
get foo
"""
    nodes = parser.parse(text)
    assert len(nodes) == 8


def test_parser_normal_payload(parser):
    text = r"""// Comment
    pUt /somewhere //here
{
    "foo": "bar", // a comment
    "hello": 1.0,
    "world": [2.0, true, null, false], // more comment
    "nested": {
        "this is it": "orly?",
        "the end": [42, 'the', 'end', 'of', 'it']
    }
}"""
    nodes = parser.parse(text)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, EsApiCallNode)
    assert n.method == 'PUT'
    assert n.path == '/somewhere'
    assert str(n) == r"""pUt /somewhere {}
{"foo":"bar","hello":1.0,"world":[2.0,true,null,false],"nested":{"this is it":"orly?","the end":[42,'the','end','of','it']}}
"""


def test_parser_file_payload(parser):
    text = """get _abc
@some_file
"""
    nodes = parser.parse(text)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, EsApiCallNode)
    assert n.method == 'GET'
    assert n.path == '/_abc'


def test_parser_string_escapes(parser):
    text = r"""geT out
{
    "'hello\tworld'": '"hello\tworld"',
    "foo\\\t\nbar": 'foo\\\t\nbar',
    "magic\\'\"": 'magic\\"\''
}"""
    nodes = parser.parse(text)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, EsApiCallNode)
    assert str(n) == r"""geT out {}
{"'hello\tworld'":'"hello\tworld"',"foo\\\t\nbar":'foo\\\t\nbar',"magic\\'\"":'magic\\"\''}
"""


def test_parser_tdqs(parser):
    text = r'''post /away
    {
        "'hello\tworld'": """"hello\t
world\"""",
        "foo\\\t\nbar": """foo\\
\t\nbar""",
        "magic\\'\"": """magic\\"\''"""
    }'''
    nodes = parser.parse(text)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, EsApiCallNode)
    assert str(n) == r'''post /away {}
{"'hello\tworld'":""""hello\t
world\"""","foo\\\t\nbar":"""foo\\
\t\nbar""","magic\\'\"":"""magic\\"\''"""}
'''


def test_parser_tsqs(parser):
    text = r"""delete it
{
        "'hello\tworld'": ''''hello\t
world\'''',
        "foo\\\t\nbar": '''foo\\
\t\nbar''',
        "magic\\'\"": '''magic\\"\'"'''
    }"""
    nodes = parser.parse(text)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, EsApiCallNode)
    assert str(n) == r"""delete it {}
{"'hello\tworld'":''''hello\t
world\'''',"foo\\\t\nbar":'''foo\\
\t\nbar''',"magic\\'\"":'''magic\\"\'"'''}
"""


def test_parser_bulk_index(parser):
    text = '''PUT _bulk
{ "index" : { "_index" : "test", "_id" : "1" } }
{ "field1" : "value1" }
{ "delete" : { "_index" : "test", "_id" : "2" } }
{ "create" : { "_index" : "test", "_id" : "3" } }
{ "field1" : "value3" }
{ "update" : {"_id" : "1", "_index" : "test"} }
{ "doc" : {"field2" : "value2"} }
'''
    nodes = parser.parse(text)
    print(nodes)
    assert len(nodes) == 1
    n = nodes[0]
    assert isinstance(n, EsApiCallNode)
    print(n)
    assert str(n) == r'''PUT _bulk {}
{"index":{"_index":"test","_id":"1"}}
{"field1":"value1"}
{"delete":{"_index":"test","_id":"2"}}
{"create":{"_index":"test","_id":"3"}}
{"field1":"value3"}
{"update":{"_id":"1","_index":"test"}}
{"doc":{"field2":"value2"}}
'''


def test_parser_invalid_missing_comma(parser):
    text = """get abc
{"a": 1 2,
 "b": 5 }"""
    with pytest.raises(PeekSyntaxError) as e:
        parser.parse(text)

    assert 'Syntax error at Line 2, Column 9' in str(e.value)

    text = """get abc
{"a": [ 3 4 ]}"""
    with pytest.raises(PeekSyntaxError) as e:
        parser.parse(text)

    assert 'Syntax error at Line 2, Column 11' in str(e.value)


def test_parser_incomplete(parser):
    text = """get
"""
    with pytest.raises(PeekSyntaxError) as e:
        parser.parse(text)

    assert 'Syntax error at Line 1, Column 4' in str(e.value)
    assert 'HTTP path must be either text literal or an expression enclosed by parenthesis' in str(e.value)


def test_parser_expr(parser):
    text = '''f 1 * 2 + (3 - 2) g(3)
'''
    nodes = parser.parse(text)
    assert 1 == len(nodes)
    assert str(nodes[0]) == "f [] [1 * 2 + (3 - 2),g [] [3] {}] {}\n"


def test_parser_dict_keys(parser):
    text = '''f {4: 2} {3 + 5: 2} {"a" + "b": c}'''
    nodes = parser.parse(text)
    assert str(nodes[0]) == 'f [] [{4:2},{3 + 5:2},{"a" + "b":c}] {}\n'


def test_implicit_variable(parser):
    text = '''f _."a".1'''
    nodes = parser.parse(text)
    assert str(nodes[0]) == 'f [] [_ . "a" . 1] {}\n'


def test_func_expr_chain(parser):
    text = '''f a.@b(1)'''
    nodes = parser.parse(text)
    assert str(nodes[0]) == 'f [] [a . b [] [1] {}] {}\n'



import json
import logging
from json import JSONDecodeError

import pygments
from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.formatted_text import PygmentsTokens, FormattedText, to_formatted_text
from prompt_toolkit.styles import style_from_pygments_cls, ConditionalStyleTransformation, \
    SwapLightAndDarkStyleTransformation

from peek.lexers import PeekStyle, PeekLexer, Heading, TipsMinor

_logger = logging.getLogger(__name__)


class Display:

    def __init__(self, app):
        self.app = app
        self.payload_lexer = PeekLexer(stack=('value',))
        self.style_transformation = ConditionalStyleTransformation(
            SwapLightAndDarkStyleTransformation(), self.app.config.as_bool('swap_colour'))

    @property
    def pretty_print(self):
        return self.app.config.as_bool('pretty_print')

    def info(self, source, header_text=''):
        if source is None:
            return
        if not self.app.batch_mode:
            self.tee_print(FormattedText([
                (PeekStyle.styles[Heading], '=== '),
                (PeekStyle.styles[TipsMinor], header_text)
            ]), style_transformation=self.style_transformation)
        if self._try_json(source):
            return
        # TODO: try more types
        self.tee_print(source, style_transformation=self.style_transformation)

    def error(self, source, header_text=''):
        if source is None:
            return
        if not self.app.batch_mode:
            self.tee_print(
                HTML('<ansired>--- </ansired>'),
                FormattedText([(PeekStyle.styles[TipsMinor], header_text)]),
                style_transformation=self.style_transformation)
        self.tee_print(source, style_transformation=self.style_transformation)

    def _try_json(self, source):
        if isinstance(source, str):
            try:
                source = json.loads(source)
            except JSONDecodeError as e:
                _logger.debug(f'Cannot decode string to json: {source!r} {e}')
                return False
        try:
            source = json.dumps(source, indent=2 if self.pretty_print else None)
            tokens = list(pygments.lex(source, lexer=self.payload_lexer))
            self.tee_print(PygmentsTokens(tokens), style=style_from_pygments_cls(PeekStyle),
                           style_transformation=self.style_transformation)
            return True
        except Exception as e:
            _logger.debug(f'Cannot render object as json: {source!r}, {e}')
            return False

    def tee_print(self, *args, **kwargs):
        print_formatted_text(*args, **kwargs)
        if self.app.capture.file() is not None:
            fragments = []
            for v in args:
                fragments.extend(to_text(v))
            print(''.join([v[1] for v in fragments]), file=self.app.capture.file())


def to_text(val):
    # Normal lists which are not instances of `FormattedText` are
    # considered plain text.
    if isinstance(val, list) and not isinstance(val, FormattedText):
        return to_formatted_text("{0}".format(val))
    return to_formatted_text(val, auto_convert=True)



from typing import NamedTuple

from pygments.token import _TokenType

HTTP_METHODS = ['get', 'post', 'put', 'delete']


class AlwaysNoneNameSpace:
    def __getattr__(self, name):
        return None


PeekToken = NamedTuple('PeekToken', [('index', int), ('ttype', _TokenType), ('value', str)])
NONE_NS = AlwaysNoneNameSpace()



import ast
import itertools
import json
import logging
import operator
import os
import sys
from numbers import Number
from subprocess import Popen

from pygments.token import Name

from peek.ast import Visitor, EsApiCallNode, DictNode, KeyValueNode, ArrayNode, NumberNode, \
    StringNode, Node, FuncCallNode, NameNode, TextNode, ShellOutNode, EsApiCallInlinePayloadNode, \
    EsApiCallFilePayloadNode, GroupNode, BinOpNode, UnaryOpNode, SymbolNode, LetNode, ForInNode
from peek.errors import PeekError
from peek.natives import EXPORTS
from peek.visitors import Ref

_logger = logging.getLogger(__name__)


def dot(left_operand, right_operand):
    if isinstance(left_operand, dict):
        if right_operand in left_operand:
            return left_operand[right_operand]
        else:
            raise PeekError(f'Value {left_operand!r} does not have key {right_operand!r}')
    elif isinstance(left_operand, list):
        if isinstance(right_operand, int):
            return left_operand[right_operand]
        else:
            raise PeekError(f'Cannot index array {left_operand!r} with non-integer value: {right_operand!r}')
    else:
        raise PeekError(f'Value {left_operand!r} must be either an array or dict')


def add(left_operand, right_operand):
    if isinstance(left_operand, str) and isinstance(right_operand, Number):
        return operator.add(left_operand, str(right_operand))
    elif isinstance(left_operand, Number) and isinstance(right_operand, str):
        return operator.add(str(left_operand), right_operand)
    else:
        return operator.add(left_operand, right_operand)


_BIN_OP_FUNCS = {
    '+': add,
    '-': operator.sub,
    '*': operator.mul,
    '/': operator.truediv,
    '%': operator.mod,
    '.': dot,
}

_UNARY_OP_FUNCS = {
    '+': operator.pos,
    '-': operator.neg,
}


class PeekVM(Visitor):

    def __init__(self, app):
        super().__init__()
        self.app = app
        self.context = {}
        self.builtins = EXPORTS
        if self.app.config.as_bool('load_extension'):
            self._load_extensions()
        self.es_api_payload_line = []
        self.func_args = []
        self.func_kwargs = {}

    @property
    def functions(self):
        return {k: v for k, v in itertools.chain(self.builtins.items(), self.context.items()) if callable(v)}

    def execute_node(self, node: Node):
        node.accept(self)

    def visit_es_api_call_node(self, node: EsApiCallNode):
        if isinstance(node.path_node, TextNode):
            path = node.path
        else:
            path_ref = Ref()
            with self.consumer(lambda v: path_ref.set(v)):
                node.path_node.accept(self)
            path = path_ref.get()
            path = path if path.startswith('/') else ('/' + path)

        options = Ref()
        self.push_consumer(lambda v: options.set(v))
        self._do_visit_dict_node(node.options_node)
        self.pop_consumer()
        options = options.get()

        if isinstance(node, EsApiCallInlinePayloadNode):
            dicts = []
            self.push_consumer(lambda v: dicts.append(v))
            for dict_node in node.dict_nodes:
                dict_node.accept(self)
            self.pop_consumer()
            lines = [json.dumps(d) for d in dicts]
            payload = ('\n'.join(lines) + '\n') if lines else None
        elif isinstance(node, EsApiCallFilePayloadNode):
            f = Ref()
            self.push_consumer(lambda v: f.set(v))
            node.file_node.accept(self)
            self.pop_consumer()
            with open(os.path.expanduser(f.get())) as ins:
                payload = ins.read()
                if not payload.endswith('\n'):
                    payload += '\n'
        else:
            raise ValueError(f'Unknown node: {node!r}')

        conn = options.pop('conn') if 'conn' in options else None
        headers = options.pop('headers') if 'headers' in options else {}
        xoid = options.pop('xoid') if 'xoid' in options else None
        if xoid:
            headers['x-opaque-id'] = str(xoid)
        runas = options.pop('runas') if 'runas' in options else None
        if runas is not None:
            headers['es-security-runas-user'] = runas
        if conn is not None:
            es_client = self.app.es_client_manager.get_client(conn)
        else:
            es_client = self.app.es_client_manager.current
        if options:
            self.app.display.error(f'Unknown options: {options}')
            return

        try:
            response = es_client.perform_request(node.method, path, payload, headers=headers if headers else None)
            self.context['_'] = _maybe_jsonify(response)
            self.app.display.info(response, header_text=self._get_header_text(conn, runas))
        except Exception as e:
            if getattr(e, 'info', None) and isinstance(getattr(e, 'status_code', None), int):
                self.context['_'] = _maybe_jsonify(e.info)
                self.app.display.info(e.info, header_text=self._get_header_text(conn, runas))
            else:
                self.app.display.error(e, header_text=self._get_header_text(conn, runas))
                _logger.exception(f'Error on ES API call: {node!r}')

    def visit_func_call_node(self, node: FuncCallNode):
        if isinstance(node.name_node, NameNode):
            func = self.get_value(node.name_node.token.value)
        else:
            func_ref = Ref()
            self.push_consumer(lambda v: func_ref.set(v))
            node.name_node.accept(self)
            self.pop_consumer()
            func = func_ref.get()
        if not callable(func):
            raise PeekError(f'{node.name_node!r} is not a callable, but {func!r}')

        func_symbols = Ref()
        self.push_consumer(lambda v: func_symbols.set(v))
        node.symbols_node.accept(self)
        self.pop_consumer()

        func_args = Ref()
        self.push_consumer(lambda v: func_args.set(v))
        node.args_node.accept(self)
        self.pop_consumer()

        for kv_node in node.kwargs_node.kv_nodes:
            assert isinstance(kv_node.key_node, NameNode), f'{kv_node.key_node!r}'
        func_kwargs = Ref()
        self.push_consumer(lambda v: func_kwargs.set(v))
        self._do_visit_dict_node(node.kwargs_node, resolve_key_name=False)
        self.pop_consumer()
        kwargs = func_kwargs.get()
        if func_symbols.get():
            kwargs['@'] = func_symbols.get()
        try:
            result = func(self.app, *func_args.get(), **kwargs)
            if node.is_stmt:
                self.app.display.info(result)
            else:
                self.consume(result)
        except Exception as e:
            _logger.exception(f'Error on invoking function: {node.name_node!r}')
            self.app.display.error(e)

    def visit_let_node(self, node: LetNode):
        for kv_node in node.assignments_node.kv_nodes:
            lhs_chain = []
            self.push_consumer(lambda v: lhs_chain.append(v))
            self._unwind_lhs(kv_node.key_node)
            self.pop_consumer()

            rhs = Ref()
            self.push_consumer(lambda v: rhs.set(v))
            kv_node.value_node.accept(self)
            self.pop_consumer()

            if len(lhs_chain) == 1:
                self.context[lhs_chain[0]] = rhs.get()
            else:
                if not lhs_chain[0] in self.context:
                    raise PeekError(f'Unknown name: {lhs_chain[0]!r}')
                lhs = self.context[lhs_chain[0]]
                for x in lhs_chain[1:-1]:
                    if isinstance(lhs, dict) or (isinstance(lhs, list) and isinstance(x, int)):
                        lhs = lhs[x]
                    else:
                        raise PeekError(f'Invalid lhs for assignment: {".".join(lhs_chain)}')

                x = lhs_chain[-1]
                if isinstance(lhs, dict) or (isinstance(lhs, list) and isinstance(x, int)):
                    lhs[x] = rhs.get()
                else:
                    raise PeekError(f'Invalid lhs for assignment: {lhs_chain}')

    def visit_shell_out_node(self, node: ShellOutNode):
        try:
            input_fd = self.app.prompt.input.fileno()
        except AttributeError:
            input_fd = sys.stdin.fileno()
        try:
            output_fd = self.app.prompt.output.fileno()
        except AttributeError:
            output_fd = sys.stdout.fileno()
        p = Popen(node.command, shell=True, stdin=input_fd, stdout=output_fd)
        p.wait()

    def visit_for_in_node(self, node: ForInNode):
        var_name = node.item.token.value
        items_ref = Ref()
        self.push_consumer(lambda v: items_ref.set(v))
        node.items.accept(self)
        self.pop_consumer()
        items = items_ref.get()
        if not isinstance(items, list):
            raise PeekError(f'For in loop must operator over a list, got {items!r}')

        for i in items:
            self.context[var_name] = i
            for n in node.suite:
                n.accept(self)

    def visit_key_value_node(self, node: KeyValueNode):
        node.key_node.accept(self)
        node.value_node.accept(self)

    def visit_name_node(self, node: NameNode):
        v = self.get_value(node.token.value)
        self.consume(v)

    def visit_symbol_node(self, node: SymbolNode):
        self.consume(node.token.value)

    def visit_string_node(self, node: StringNode):
        self.consume(ast.literal_eval(node.token.value))

    def visit_number_node(self, node: NumberNode):
        self.consume(ast.literal_eval(node.token.value))

    def visit_dict_node(self, node: DictNode):
        self._do_visit_dict_node(node, resolve_key_name=True)

    def visit_array_node(self, node: ArrayNode):
        values = []
        self.push_consumer(lambda v: values.append(v))
        for node in node.value_nodes:
            node.accept(self)
        self.pop_consumer()
        self.consume(values)

    def visit_text_node(self, node: TextNode):
        if node.token.ttype is Name.Builtin:
            self.consume({'true': True, 'false': False, 'null': None}[node.token.value])
        else:
            self.consume(node.token.value)

    def visit_bin_op_node(self, node: BinOpNode):
        left_operand = Ref()
        self.push_consumer(lambda v: left_operand.set(v))
        node.left_node.accept(self)
        self.pop_consumer()

        right_operand = Ref()
        self.push_consumer(lambda v: right_operand.set(v))
        node.right_node.accept(self)
        self.pop_consumer()

        op_func = _BIN_OP_FUNCS.get(node.op_token.value, None)
        if op_func is None:
            raise PeekError(f'Unknown binary operator: {node.op_token.value!r}')
        self.consume(op_func(left_operand.get(), right_operand.get()))

    def visit_unary_op_node(self, node: UnaryOpNode):
        operand = Ref()
        self.push_consumer(lambda v: operand.set(v))
        node.operand_node.accept(self)
        self.pop_consumer()

        op_func = _UNARY_OP_FUNCS.get(node.op_token.value, None)
        if op_func is None:
            raise PeekError(f'Unknown unary operator: {node.op_token.value!r}')
        self.consume(op_func(operand.get()))

    def visit_group_node(self, node: GroupNode):
        node.grouped.accept(self)

    def _do_visit_dict_node(self, node: DictNode, resolve_key_name=False):
        assert isinstance(node, DictNode)
        keys = []
        values = []
        self.push_consumer(lambda v: keys.append(v))
        for kv_node in node.kv_nodes:
            if resolve_key_name or not isinstance(kv_node.key_node, NameNode):
                kv_node.key_node.accept(self)
            else:
                self.consume(kv_node.key_node.token.value)
            self.push_consumer(lambda v: values.append(v))
            kv_node.value_node.accept(self)
            self.pop_consumer()
        self.pop_consumer()
        assert len(keys) == len(values), f'{keys!r}, {values!r}'
        self.consume(dict(zip(keys, values)))

    def get_value(self, name):
        value = self.builtins.get(name)
        if value is None:
            value = self.context.get(name)
        if value is None:
            raise PeekError(f'Unknown name: {name!r}')
        return value

    def _unwind_lhs(self, node: Node):
        if isinstance(node, NameNode):
            self.consume(node.token.value)
        elif isinstance(node, BinOpNode):
            if node.op_token.value != '.':
                raise PeekError(f'lhs can only have variable and dot notation, but got {node!r}')
            self._unwind_lhs(node.left_node)
            node.right_node.accept(self)
        else:
            raise PeekError(f'lhs can only have variable and dot notation, but got {node!r}')

    def _load_extensions(self):
        """
        Load extra variables from external paths
        """
        extension_path = self.app.config['extension_path']
        if not extension_path:
            return

        sys_path = sys.path[:]
        try:
            for p in extension_path.split(':'):
                if os.path.isfile(p):
                    self._load_one_extension(p)
                elif os.path.isdir(p):
                    for f in os.listdir(p):
                        if not f.endswith('.py'):
                            continue
                        self._load_one_extension(os.path.join(p, f))
                else:
                    _logger.warning(f'Cannot load extension path: {p}')
        finally:
            sys.path = sys_path

    def _load_one_extension(self, p):
        _logger.debug(f'Loading extension: {p!r}')
        import importlib
        fields = os.path.splitext(p)
        if len(fields) != 2 or fields[1] != '.py':
            _logger.warning(f'Extension must be python files, got: {p!r}')
            return
        sys.path.insert(0, os.path.dirname(fields[0]))
        try:
            m = importlib.import_module(os.path.basename(fields[0]))
            exports = getattr(m, 'EXPORTS', None)
            if isinstance(exports, dict):
                self.context.update(exports)
                _logger.info(f'Loaded extension: {p!r}')
            else:
                _logger.warning(f'Ignore extension {p!r} since EXPORTS is not a dict, but: {exports!r}')
        except Exception as e:
            _logger.error(f'Error on loading extension: {p!r}, {e}')
            _logger.exception(f'Error on loading extension: {p!r}')

    def _get_header_text(self, conn, runas):
        parts = []
        if conn is not None:
            parts.append(f'conn={conn!r}')
        if runas is not None:
            parts.append(f'runas={runas!r}')
        return ' '.join(parts)


def _maybe_jsonify(r):
    try:
        return json.loads(r)
    except Exception:
        return r



import ast
import json
import logging
from typing import Iterable

from pygments.token import Token, Whitespace, String, Comment, Literal, Number, Name, Error

from peek.ast import NameNode, FuncCallNode, KeyValueNode, StringNode, NumberNode, TextNode, DictNode, \
    ArrayNode, ShellOutNode, EsApiCallInlinePayloadNode, EsApiCallFilePayloadNode, BinOpNode, UnaryOpNode, GroupNode, \
    SymbolNode, LetNode, ForInNode
from peek.common import PeekToken
from peek.errors import PeekSyntaxError
from peek.lexers import PeekLexer, BlankLine, CurlyLeft, DictKey, Colon, \
    CurlyRight, Comma, BracketLeft, BracketRight, TripleS, TripleD, EOF, FuncName, Assign, HttpMethod, OptionName, \
    ShellOut, At, ParenLeft, ParenRight, UnaryOp, BinOp, Let, For, In

_logger = logging.getLogger(__name__)

HTTP_METHODS = ['GET', 'PUT', 'POST', 'DELETE']

_BIN_OP_ORDERS = {
    None: -1,
    '+': 100,
    '-': 100,
    '*': 200,
    '/': 200,
    '%': 200,
    '.': 300,
}


class PeekParser:

    def __init__(self):
        self.lexer = PeekLexer()
        self.text = ''
        self.position = 0
        self.tokens = []

    def parse(self, text):
        self.text = text
        self.position = 0
        self.tokens = []

        self.tokens = process_tokens(self.lexer.get_tokens_unprocessed(self.text))
        for token in self.tokens:
            if token.ttype in Token.Error:
                raise PeekSyntaxError(self.text, token)

        return self._do_parse()

    def _do_parse(self):
        nodes = []
        while self._peek_token().ttype is not EOF:
            token = self._peek_token()
            if token.ttype is BlankLine:
                self._consume_token(BlankLine)
            else:
                nodes.append(self._parse_stmt())
        return nodes

    def _parse_stmt(self):
        token = self._peek_token()
        if token.ttype is FuncName:
            return self._parse_func_call()
        elif token.ttype is HttpMethod:
            return self._parse_es_api_call()
        elif token.ttype is Let:
            return self._parse_let_stmt()
        elif token.ttype is ShellOut:
            return self._parse_shell_out()
        elif token.ttype is For:
            return self._parse_for_stmt()
        else:
            raise PeekSyntaxError(
                self.text, token,
                title='Invalid token',
                message='Expect beginning of a statement')

    def _parse_es_api_call(self):
        method_token = self._consume_token(HttpMethod)
        method_node = NameNode(method_token)
        if method_token.value.upper() not in HTTP_METHODS:
            raise PeekSyntaxError(
                self.text, method_token,
                title='Invalid HTTP method',
                message=f'Expect HTTP method of value in {HTTP_METHODS!r}, got {method_token.value!r}')
        if self._peek_token().ttype is Literal:
            path_node = TextNode(self._consume_token(Literal))
        elif self._peek_token().ttype is ParenLeft:
            path_node = self._parse_expr()
        else:
            raise PeekSyntaxError(
                self.text, self._peek_token(),
                message='HTTP path must be either text literal or an expression enclosed by parenthesis')
        option_nodes = []
        while self._peek_token().ttype is OptionName:
            n = NameNode(self._consume_token(OptionName))
            self._consume_token(Assign)
            option_nodes.append(KeyValueNode(n, self._parse_expr()))

        if self._peek_token().ttype is At:
            self._consume_token(At)
            return EsApiCallFilePayloadNode(
                method_node, path_node, DictNode(option_nodes), TextNode(self._consume_token(Literal)))
        else:
            dict_nodes = []
            while self._peek_token().ttype is not EOF:
                if self._peek_token().ttype is CurlyLeft:
                    dict_nodes.append(self._parse_dict())
                else:
                    break
            return EsApiCallInlinePayloadNode(method_node, path_node, DictNode(option_nodes), dict_nodes)

    def _parse_let_stmt(self):
        self._consume_token(Let)
        kv_nodes = []
        while self._peek_token().ttype is not EOF:
            if self._peek_token().ttype is BlankLine:
                self._consume_token(BlankLine)
                break
            key_node = self._parse_expr()
            self._consume_token(Assign)
            value_node = self._parse_expr()
            kv_nodes.append(KeyValueNode(key_node, value_node))
        return LetNode(DictNode(kv_nodes))

    def _parse_for_stmt(self):
        self._consume_token(For)
        item = NameNode(self._consume_token(Name))
        self._consume_token(In)
        items = self._parse_expr()
        self._consume_token(CurlyLeft)
        suite = []
        while self._peek_token().ttype is not CurlyRight:
            token = self._peek_token()
            if token.ttype is BlankLine:
                self._consume_token(BlankLine)
            else:
                suite.append(self._parse_stmt())
        self._consume_token(CurlyRight)
        return ForInNode(item, items, suite)

    def _parse_dict(self):
        kv_nodes = []
        self._consume_token(CurlyLeft)
        while self._peek_token().ttype is not CurlyRight:
            kv_nodes.append(self._parse_dict_kv())
            if self._peek_token().ttype is Comma:
                self._consume_token(Comma)
            else:
                break
        self._consume_token(CurlyRight)
        return DictNode(kv_nodes)

    def _parse_dict_kv(self):
        if self._peek_token().ttype is DictKey:
            key_node = StringNode(self._consume_token(DictKey))
        else:
            key_node = self._parse_expr()
        self._consume_token(Colon)
        value_node = self._parse_expr()
        return KeyValueNode(key_node, value_node)

    def _parse_array(self):
        value_nodes = []
        self._consume_token(BracketLeft)
        while self._peek_token().ttype is not BracketRight:
            value_nodes.append(self._parse_expr())
            if self._peek_token().ttype is Comma:
                self._consume_token(Comma)
            else:
                break
        self._consume_token(BracketRight)
        return ArrayNode(value_nodes)

    def _parse_value(self):
        token = self._peek_token()
        if token.ttype is CurlyLeft:
            return self._parse_dict()
        elif token.ttype is BracketLeft:
            return self._parse_array()
        elif token.ttype in Number:
            return NumberNode(self._consume_token(token.ttype))
        elif token.ttype is Name.Builtin:
            return TextNode(self._consume_token(token.ttype))
        elif token.ttype in (String.Double, String.Single, TripleS, TripleD):
            return StringNode(self._consume_token(token.ttype))
        else:
            raise PeekSyntaxError(self.text, token, message=f'Unexpected token: {token!r}')

    def _parse_func_call(self):
        name_node = NameNode(self._consume_token(FuncName))
        symbol_nodes, arg_nodes, kwarg_nodes = self._parse_func_call_args()
        return FuncCallNode(name_node, ArrayNode(symbol_nodes), ArrayNode(arg_nodes), DictNode(kwarg_nodes))

    def _parse_func_call_args(self, is_stmt=True):
        symbol_nodes = []
        arg_nodes = []
        kwarg_nodes = []
        while self._peek_token().ttype is not EOF:
            if self._peek_token().ttype is BlankLine:
                self._consume_token(BlankLine)
                if is_stmt:
                    break
            elif self._peek_token().ttype is ParenRight:
                if is_stmt:
                    raise PeekSyntaxError(
                        self.text, self._peek_token(),
                        message='Found function expression while parsing for function stmt')
                else:
                    break
            elif self._peek_token().ttype is Name:
                n = NameNode(self._consume_token(Name))
                if self._peek_token().ttype is Assign:
                    self._consume_token(Assign)
                    kwarg_nodes.append(KeyValueNode(n, self._parse_expr()))
                elif self._peek_token().ttype is ParenLeft:  # nested function expr
                    self._consume_token(ParenLeft)
                    sub_symbol_nodes, sub_arg_nodes, sub_kwarg_nodes = self._parse_func_call_args(is_stmt=False)
                    arg_nodes.append(FuncCallNode(
                        n,
                        ArrayNode(sub_symbol_nodes),
                        ArrayNode(sub_arg_nodes),
                        DictNode(sub_kwarg_nodes),
                        is_stmt=False
                    ))
                    self._consume_token(ParenRight)
                else:
                    arg_nodes.append(self._parse_expr_after_left_operand(n))
            elif self._peek_token().ttype is At:
                self._consume_token(At)
                symbol_nodes.append(SymbolNode(self._consume_token(Literal)))
            else:
                arg_nodes.append(self._parse_expr())
        return symbol_nodes, arg_nodes, kwarg_nodes

    def _parse_shell_out(self):
        self._consume_token(ShellOut)
        return ShellOutNode(TextNode(self._consume_token(Literal)))

    def _parse_expr(self, last_bin_op=None):
        if self._peek_token().ttype is UnaryOp:
            unary_op_token = self._consume_token(UnaryOp)
        else:
            unary_op_token = None

        if self._peek_token().ttype is ParenLeft:
            pl = self._consume_token(ParenLeft)
            n = self._parse_expr()
            pr = self._consume_token(ParenRight)
            n = GroupNode(n, pl, pr)
        elif self._peek_token().ttype is Name:
            n = NameNode(self._consume_token(Name))
        elif self._peek_token().ttype is At:
            self._consume_token(At)
            n = SymbolNode(self._consume_token(Literal))
        else:
            n = self._parse_value()

        return self._parse_expr_after_left_operand(n, unary_op_token=unary_op_token, last_bin_op=last_bin_op)

    def _parse_expr_after_left_operand(self, n, unary_op_token=None, last_bin_op=None):
        while True:
            if self._peek_token().ttype is BinOp:
                op_token = self._peek_token()
                if _BIN_OP_ORDERS[last_bin_op] >= _BIN_OP_ORDERS[op_token.value]:
                    return n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
                else:
                    self._consume_token(op_token.ttype)
                    right_node = self._parse_expr(last_bin_op=op_token.value)
                    n = n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
                    n = BinOpNode(op_token, n, right_node)
            elif self._peek_token().ttype is ParenLeft:  # func call
                if last_bin_op == '.':
                    return n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
                else:
                    self._consume_token(ParenLeft)
                    symbol_nodes, arg_nodes, kwarg_nodes = self._parse_func_call_args(is_stmt=False)
                    self._consume_token(ParenRight)
                    n = FuncCallNode(
                        n,
                        ArrayNode(symbol_nodes),
                        ArrayNode(arg_nodes),
                        DictNode(kwarg_nodes),
                        is_stmt=False
                    )
                    n = n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
            else:
                return n if unary_op_token is None else UnaryOpNode(unary_op_token, n)
            unary_op_token = None

    def _peek_token(self) -> PeekToken:
        if self.position >= len(self.tokens):
            return PeekToken(len(self.text), EOF, '\0')
        return self.tokens[self.position]

    def _consume_token(self, ttype, value=None) -> PeekToken:
        token = self._peek_token()
        self.position += 1
        if token.ttype is not ttype:
            raise PeekSyntaxError(self.text, token,
                                  message=f'Expect token of type {ttype!r}, got {token.ttype!r}')
        if value and (token.value != value or token.value not in value):
            raise PeekSyntaxError(self.text, token,
                                  message=f'Expect token of value {value!r}, got {token.value!r}')
        return token


def normalise_string(value):
    return json.dumps(ast.literal_eval(value))


def process_tokens(tokens: Iterable[PeekToken]):
    """
    Process tokens by filtering out whitespaces and merging tokens that
    should be represented as one, e.g. Strings, BlankLine, Error.
    """
    processed_tokens = []
    current_token = None
    for token in tokens:
        if token.ttype in (Whitespace, Comment.Single):
            if current_token is not None:
                processed_tokens.append(current_token)
                current_token = None
        elif token.ttype in String or token.ttype is BlankLine or token.ttype is Error:
            if current_token is not None and current_token.ttype is token.ttype:
                current_token = PeekToken(
                    current_token.index, current_token.ttype, current_token.value + token.value)
            else:
                # two consecutive strings with different quotes should not be merged
                if current_token is not None:
                    processed_tokens.append(current_token)
                current_token = token
        else:
            if current_token is not None:
                processed_tokens.append(current_token)
                current_token = None
            processed_tokens.append(token)
    if current_token is not None:
        processed_tokens.append(current_token)

    _logger.debug(f'processed tokens: {processed_tokens}')
    return processed_tokens



#!/usr/bin/env python

"""The setup script."""
import os
import re

from setuptools import setup, find_packages


def get_version():
    with open(os.path.join(os.path.dirname(__file__), 'peek', '__init__.py')) as ins:
        return re.search("__version__ = ['\"]([^'\"]+)['\"]", ins.read()).group(1)


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as requirements_file:
    requirements = [line.strip() for line in requirements_file.read().splitlines() if line.strip()]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Yang Wang",
    author_email='ywangd@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Peek into elasticsearch clusters",
    entry_points={
        'console_scripts': [
            'peek=peek.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='peek,elasticsearch,cli',
    name='es-peek',
    packages=find_packages(include=['peek', 'peek.*']),
    setup_requires=setup_requirements,
    extras_require={
        'full': ['kerberos~=1.3.0', 'pyperclip~=1.8.0']
    },
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/ywangd/peek',
    version=get_version(),
    zip_safe=False,
)



import os
from typing import Iterable
from unittest.mock import MagicMock

from prompt_toolkit.completion import CompleteEvent, Completion
from prompt_toolkit.document import Document
from pygments.token import Literal

from peek.common import PeekToken
from peek.completer import load_specs, PeekCompleter, find_beginning_token, matchable_specs
from peek.lexers import HttpMethod, FuncName, BlankLine
from peek.natives import EXPORTS

mock_app = MagicMock(name='PeekApp')
mock_app.vm.functions = {k: v for k, v in EXPORTS.items() if callable(v)}
mock_app.config.as_bool.return_value = True

completer = PeekCompleter(mock_app)


def equivalent_completions(c0: Completion, c1: Completion):
    return c0.text == c1.text and c0.start_position == c1.start_position


def completions_has(cs: Iterable[Completion], *cc: Completion):
    if not completer.specs:
        return True

    actual = set((x.text, x.start_position) for x in cs)
    expected = set((x.text, x.start_position) for x in cc)
    if len(expected) == 0:
        ret = len(actual) == 0
    else:
        ret = actual.issuperset(expected)
    if ret is False:
        print(f'actual: {actual!r} is not superset of {expected!r}')
    return ret


def get_completions(document: Document):
    return completer.get_completions(document, CompleteEvent(True))


def test_load_specs():
    from peek import __file__ as package_root
    package_root = os.path.dirname(package_root)
    kibana_dir = os.path.join(package_root, 'specs', 'kibana-7.8.1')
    specs = load_specs(kibana_dir)
    # Skip the test if no specs are found
    if not specs:
        return

    # Make sure loading and merging work
    print(specs['indices.create']['data_autocomplete_rules'])
    print(specs['security.put_user']['data_autocomplete_rules'])

    next(matchable_specs('POST', ['_security', 'api_key'], specs))

    next(matchable_specs('POST', ['_security', 'oauth2', 'token'], specs,
                         required_field='data_autocomplete_rules'))


def test_find_beginning_token():
    tokens = [
        PeekToken(index=0, ttype=HttpMethod, value='get'),
        PeekToken(index=4, ttype=Literal, value='abc'),
        PeekToken(index=8, ttype=FuncName, value='ge')
    ]
    i, t = find_beginning_token(tokens)
    assert i == 2

    tokens = [
        PeekToken(index=0, ttype=FuncName, value='connect'),
        PeekToken(index=7, ttype=BlankLine, value='\n'),
        PeekToken(index=8, ttype=FuncName, value='session')
    ]
    i, t = find_beginning_token(tokens)
    assert i == 2


def test_complete_http_method_and_func_name():
    assert completions_has(
        get_completions(Document('po')),
        Completion(text='POST', start_position=-2),
    )

    assert completions_has(
        get_completions(Document('''get abc
ge''')),
        Completion(text='GET', start_position=-2),
    )

    assert completions_has(
        get_completions(Document('con')),
        Completion(text='config', start_position=-3),
        Completion(text='connect', start_position=-3),
    )

    assert completions_has(
        get_completions(Document('''config
sa''')),
        Completion(text='saml_authenticate', start_position=-2),
    )


def test_complete_func_option_name():
    assert completions_has(
        get_completions(Document('connect ')),
        Completion(text='hosts=', start_position=0),
        Completion(text='api_key=', start_position=0),
    )

    assert completions_has(
        get_completions(Document('''connect
session ''')),
        Completion(text='current=', start_position=0),
        Completion(text='remove=', start_position=0),
        Completion(text='@info', start_position=0),
    )


def test_complete_func_option_name_with_partial():
    assert completions_has(
        get_completions(Document('connect user')),
        Completion(text='username=', start_position=-4),
    )

    assert completions_has(
        get_completions(Document('''config
saml_authenticate r''')),
        Completion(text='realm=', start_position=-1),
    )

    assert completions_has(
        get_completions(Document('session @r')),
        Completion(text='@remove', start_position=-2),
    )


def test_complete_option_name_should_not_be_in_value_place():
    assert completions_has(
        get_completions(Document('''connect hosts='''))
    )

    assert completions_has(
        get_completions(Document('''connect hosts=h'''))
    )

    assert completions_has(
        get_completions(Document('''get _search runas='''))
    )

    assert completions_has(
        get_completions(Document('''get _search runas=r'''))
    )


def test_complete_http_path():
    assert completions_has(
        get_completions(Document('get _secapi')),
        Completion(text='_security/api_key', start_position=-7),
        Completion(text='_security/role_mapping', start_position=-7),
    )

    assert completions_has(
        get_completions(Document('get _security/api_key?')),
        Completion(text='name', start_position=0),
        Completion(text='realm_name', start_position=0),
    )

    assert completions_has(
        get_completions(Document('get _security/api_key?o')),
        Completion(text='owner', start_position=-1),
    )

    assert completions_has(
        get_completions(Document('get _security/api_key?owner=')),
        Completion(text='true', start_position=0),
    )

    assert completions_has(
        get_completions(Document('post _security/api_key?refresh=')),
        Completion(text='wait_for', start_position=0),
    )

    assert completions_has(
        get_completions(Document('''get _security/api_key
post token''')),
        Completion(text='_security/oauth2/token', start_position=-5),
    )


def test_complete_http_options():
    assert completions_has(
        get_completions(Document('get _search ')),
        Completion(text='conn=', start_position=0),
        Completion(text='runas=', start_position=0),
    )

    assert completions_has(
        get_completions(Document('get _search r')),
        Completion(text='runas=', start_position=-1),
    )

    assert completions_has(
        get_completions(Document('''get _search
get _cluster/health c''')),
        Completion(text='conn=', start_position=-1),
    )

    assert len(list(get_completions(Document('''post _security/api_key
{"role_descriptors": }''', 44)))) == 0


def test_not_complete_http_options():
    assert len(list(get_completions(Document('''POST _security/oauth2/token
''')))) == 0


def test_not_complete_http_path():
    assert completions_has(
        get_completions(Document('''get
''')))


def test_payload_completion_000():
    assert completions_has(
        get_completions(Document('''POST _security/api_key
{""}''', 25)),
        Completion(text="role_descriptors", start_position=0),
    )


def test_payload_completion_001():
    assert completions_has(
        get_completions(Document('''POST _security/api_key
{
  "role_descriptors": {
    "role_name": {
      "indices": [
        {""}
      ]
    }
  }
}''', 97)),
        Completion(text="field_security", start_position=0),
    )


def test_payload_completion_002():
    assert completions_has(
        get_completions(Document('''POST _security/api_key
{
  "role_descriptors": {
    "role_name": {
      "indices": [
        {"field_security": ""}
      ]
    }
  },
  "n"
}''', 141)),
        Completion(text="name", start_position=-1),
    )


def test_payload_completion_003():
    assert completions_has(
        get_completions(Document('''POST _security/api_key
{
  "role_descriptors": {
    "role_name": {
      "cluster": [],
      ""
    }
  }
}''', 97)),
        Completion(text="indices"),
    )


def test_payload_completion_004():
    assert completions_has(
        get_completions(Document('''POST _security/oauth2/token
{""}''', 30)),
        Completion(text="scope", start_position=0),
    )


def test_file_payload_completion():
    f = os.listdir('.')[0]
    assert completions_has(
        get_completions(Document('''get /
@''')),
        Completion(text=f, start_position=0)
    )



import json
import logging
import os
from abc import ABCMeta, abstractmethod
from typing import List

from configobj import Section
from elasticsearch import Elasticsearch, AuthenticationException
from urllib3 import Timeout

from peek.errors import PeekError

_logger = logging.getLogger(__name__)


class NoopDeserializer:
    def __init__(self):
        pass

    def loads(self, s, *args, **kwargs):
        return s


noopDeserializer = NoopDeserializer()


class BaseClient(metaclass=ABCMeta):

    @abstractmethod
    def perform_request(self, method, path, payload=None, deserialize_it=False, **kwargs):
        pass


class EsClient(BaseClient):

    def __init__(self,
                 name=None,
                 hosts=None,
                 cloud_id=None,
                 username=None,
                 password=None,
                 use_ssl=False,
                 verify_certs=False,
                 assert_hostname=False,
                 ca_certs=None,
                 client_cert=None,
                 client_key=None,
                 api_key=None,
                 token=None,
                 headers=None,
                 **kwargs):

        self.name = name
        self.hosts = hosts
        self.cloud_id = cloud_id
        self.auth = f'{username}:{password}' if username and password else None
        self.use_ssl = use_ssl
        self.verify_certs = verify_certs
        self.assert_hostname = assert_hostname
        self.ca_certs = ca_certs
        self.client_cert = client_cert
        self.client_key = client_key
        self.api_key_id = api_key[0] if api_key else None
        self.token = token
        self.headers = dict(headers) if headers is not None else None
        if token:
            token_header = {'Authorization': f'Bearer {token}'}
            if headers:
                headers.update(token_header)
            else:
                headers = token_header

        self.es = Elasticsearch(
            hosts=self.hosts.split(',') if self.hosts else None,
            cloud_id=cloud_id,
            http_auth=self.auth,
            use_ssl=use_ssl,
            verify_certs=verify_certs,
            ca_certs=ca_certs,
            client_cert=client_cert,
            client_key=client_key,
            ssl_show_warn=False,
            timeout=Timeout(connect=None, read=None),
            api_key=api_key,
            headers=headers,
            ssl_assert_hostname=assert_hostname,
            **kwargs,
        )

    def perform_request(self, method, path, payload=None, deserialize_it=False, **kwargs):
        _logger.debug(f'Performing request: {method!r}, {path!r}, {payload!r}')
        deserializer = self.es.transport.deserializer
        try:
            if not deserialize_it:
                # Avoid deserializing the response since we parse it with the main loop for syntax highlighting
                self.es.transport.deserializer = noopDeserializer
            return self.es.transport.perform_request(method, path, body=payload, **kwargs)
        finally:
            if not deserialize_it:
                self.es.transport.deserializer = deserializer

    def info(self):
        if self.api_key_id:
            auth = f'ApiKey {self.api_key_id[:10]}...'
        elif self.token:
            auth = f'Token {self.token[:10]}...'
        elif self.auth:
            auth = f'Username {self.auth.split(":")[0]}'
        else:
            auth = None

        if self.headers:
            headers = list(self.headers.keys())
        else:
            headers = None

        return {
            'name': self.name,
            'hosts': self.hosts,
            'cloud_id': self.cloud_id,
            'auth': auth,
            'use_ssl': self.use_ssl,
            'verify_certs': self.verify_certs,
            'ca_certs': self.ca_certs,
            'client_cert': self.client_cert,
            'client_key': self.client_key,
            'headers': headers,
        }

    def __str__(self):
        if self.name:
            return f'{self.name}'

        hosts = []
        if self.hosts:
            for host in self.hosts.split(','):
                if host.startswith('https://') or host.startswith('http://'):
                    hosts.append(host)
                else:
                    hosts.append(('https://' if self.use_ssl else 'http://') + host)
        else:
            hosts.append(self.cloud_id)

        hosts = ','.join(hosts)
        if self.api_key_id:
            return f'K-{self.api_key_id[:10]} @ {hosts}'
        elif self.token:
            return f'T-{self.token[:10]} @ {hosts}'
        elif self.auth:
            username = self.auth.split(':')[0]
            return f'{username} @ {hosts}'
        else:
            return f'{hosts}'


class RefreshingEsClient(BaseClient):
    # TODO: Given a pair of access_token and refresh token, refresh to keep login

    def __init__(self,
                 parent: EsClient,
                 username,
                 access_token,
                 refresh_token,
                 expires_in,
                 name=None):

        self.parent = parent
        self.username = username
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires_in = expires_in
        self.name = name
        self.delegate = self._build_delegate()

    def __getattr__(self, item):
        return getattr(self.delegate, item)

    def perform_request(self, method, path, payload=None, deserialize_it=False, **kwargs):
        try:
            return self.delegate.perform_request(method, path, payload, deserialize_it, **kwargs)
        except AuthenticationException as e:
            if e.status_code == 401:
                response = self.parent.perform_request(
                    'POST', '/_security/oauth2/token',
                    json.dumps({
                        'grant_type': 'refresh_token',
                        'refresh_token': self.refresh_token,
                    }),
                    deserialize_it=True)
                self.access_token = response['access_token']
                self.refresh_token = response['refresh_token']
                self.expires_in = response['expires_in']
                self._build_delegate()
                self.perform_request(method, path, payload, deserialize_it=deserialize_it, **kwargs)

    def info(self):
        info = self.delegate.info()
        info['username'] = self.username
        return info

    def __str__(self):
        if self.name:
            return f'{self.name}'
        else:
            return f'{self.username} @ {self.delegate}'

    def _build_delegate(self):
        return EsClient(
            name=self.name,
            hosts=self.parent.hosts,
            cloud_id=self.parent.cloud_id,
            use_ssl=self.parent.use_ssl,
            verify_certs=self.parent.verify_certs,
            ca_certs=self.parent.ca_certs,
            client_cert=self.parent.client_cert,
            client_key=self.parent.client_key,
            headers=self.parent.headers,
            token=self.access_token,
        )


class EsClientManager:

    def __init__(self):
        self._clients: List[EsClient] = []
        self._index_current = None

    def add(self, client):
        self._clients.append(client)
        self._index_current = len(self._clients) - 1
        # TODO: maintain size

    @property
    def current(self):
        if self._index_current is None:
            raise PeekError('No ES client is configured')
        if self._index_current < 0 or self._index_current >= len(self._clients):
            raise PeekError(f'Attempt to get ES client at invalid index [{self._index_current}]')
        return self._clients[self._index_current]

    def set_current(self, x):
        if isinstance(x, str):
            name = x
            self._index_current = self._clients.index(self.get_client(name))
        elif isinstance(x, int):
            if x < 0 or x >= len(self._clients):
                raise PeekError(f'Attempt to set ES client at invalid index [{x}]')
            self._index_current = x
        else:
            raise ValueError(f'Connection must be specified by either name or index, got {x!r}')

    def clients(self):
        return self._clients

    def get_client(self, x=None):
        if x is None:
            return self.current
        elif isinstance(x, str):
            name = x
            if not name:
                raise ValueError('Name cannot be empty')
            for c in self._clients:
                if c.name == name:
                    return c
            else:
                raise ValueError(f'No client with name: {name!r}')
        elif isinstance(x, int):
            if x < 0 or x >= len(self._clients):
                raise PeekError(f'Attempt to set ES client at invalid index [{x}]')
            return self._clients[x]
        else:
            raise ValueError(f'Connection must be specified by either name or index, got {x!r}')

    def remove_client(self, x=None):
        if x is None:
            self.remove_client(self._index_current)
        elif isinstance(x, str):
            idx = self._clients.index(self.get_client(x))
            self.remove_client(idx)
        elif isinstance(x, int):
            if len(self._clients) == 1:
                raise PeekError('Cannot delete the last connection')
            if x < 0 or x >= len(self._clients):
                raise PeekError(f'Attempt to remove ES client at invalid index [{x}]')
            self._clients.pop(x)
            if not self._clients:
                self._index_current = None
                return
            if x < self._index_current:
                self._index_current -= 1
            elif x == self._index_current:
                self._index_current = 0
        else:
            raise ValueError(f'Connection must be specified by either name or index, got {x!r}')

    def __str__(self):
        lines = []
        for i, client in enumerate(self.clients()):
            prefix = '*' if client == self.current else ' '
            index = f'[{i}]'
            lines.append(f'{prefix} {index:>4} {client}')
        return '\n'.join(lines)


DEFAULT_OPTIONS = {
    'name': None,
    'hosts': 'localhost:9200',
    'cloud_id': None,
    'username': None,
    'password': None,
    'api_key': None,
    'token': None,
    'use_ssl': None,
    'verify_certs': False,
    'assert_hostname': False,
    'ca_certs': None,
    'client_cert': None,
    'client_key': None,
    'headers': None,
    'force_prompt': False,
    'no_prompt': False,
}


def connect(app, **options):
    final_options = dict(DEFAULT_OPTIONS)

    if isinstance(app.config.get('connection'), Section):
        final_options.update({k: v for k, v in app.config.get('connection').dict().items() if v})

    # Override with provided options, including null values since it could be intentional
    final_options.update({k: v for k, v in options.items()})

    if final_options['cloud_id'] is not None:
        final_options['hosts'] = None

    if final_options['api_key']:
        return _connect_api_key(app, **final_options)
    elif final_options['token']:
        return _connect_token(app, **final_options)
    else:
        return _connect_userpass(app, **final_options)


def _connect_userpass(app, **options):
    username = options.get('username')
    password = options.get('password')
    if options['hosts']:
        service_name = f'peek/{options["hosts"]}/userpass'
    else:
        service_name = f'peek/{options["cloud_id"]}/userpass'

    if not username and password:
        raise PeekError('Username is required for userpass authentication')

    if options['force_prompt']:
        password = app.input(message='Please enter password: ', is_secret=True)

    if username and not password:
        password = os.environ.get('PEEK_PASSWORD', None)
        if not password:
            if app.config.as_bool('use_keyring'):
                password = _keyring(service_name, username)
                if not password:
                    if options['no_prompt']:
                        raise PeekError('Password is not found and password prompt is disabled')
                    password = app.input(message='Please enter password: ', is_secret=True)

            else:
                if options['no_prompt']:
                    raise PeekError('Password is not found and password prompt is disabled')
                password = app.input(message='Please enter password: ', is_secret=True)

    if username and password and app.config.as_bool('use_keyring'):
        _keyring(service_name, username, password)

    return EsClient(
        name=options['name'],
        hosts=options['hosts'],
        cloud_id=options['cloud_id'],
        username=username,
        password=password,
        use_ssl=options['use_ssl'],
        verify_certs=options['verify_certs'],
        assert_hostname=options['assert_hostname'],
        ca_certs=options['ca_certs'],
        client_cert=options['client_cert'],
        client_key=options['client_key'],
        headers=options['headers'],
    )


def _connect_api_key(app, **options):
    _logger.debug('Connecting with API key')
    return EsClient(
        name=options['name'],
        hosts=options['hosts'],
        cloud_id=options['cloud_id'],
        api_key=options['api_key'].split(':'),
        use_ssl=options['use_ssl'],
        verify_certs=options['verify_certs'],
        assert_hostname=options['assert_hostname'],
        ca_certs=options['ca_certs'],
        client_cert=options['client_cert'],
        client_key=options['client_key'],
        headers=options['headers'],
    )


def _connect_token(app, **options):
    _logger.debug('Connecting with token')
    return EsClient(
        name=options['name'],
        hosts=options['hosts'],
        cloud_id=options['cloud_id'],
        token=options['token'],
        use_ssl=options['use_ssl'],
        verify_certs=options['verify_certs'],
        assert_hostname=options['assert_hostname'],
        ca_certs=options['ca_certs'],
        client_cert=options['client_cert'],
        client_key=options['client_key'],
        headers=options['headers'],
    )


KEYRING = None


def _keyring(service_name, key, value=None):
    """
    When a None value is passed in, means getting value out of the keyring
    """
    global KEYRING
    if KEYRING is None:
        import importlib
        try:
            KEYRING = importlib.import_module("keyring")
        except ModuleNotFoundError as e:
            _logger.warning("import keyring failed: %r.", e)

    if KEYRING is not None:
        if value is None:
            return KEYRING.get_password(service_name, key)
        else:
            KEYRING.set_password(service_name, key, value)


class ConnectFunc:
    def __call__(self, app, **options):
        app.es_client_manager.add(connect(app, **options))
        return str(app.es_client_manager)

    @property
    def options(self):
        return dict(DEFAULT_OPTIONS)
